# -*- coding: utf-8 -*-

import jax.numpy as jn

import brainpy.math.jax as bm
from brainpy.base.base import Base
from brainpy.base.collector import ArrayCollector

__all__ = [
  # optimizers
  'Optimizer',
  'SGD',
  'Momentum',
  'NesterovMomentum',
  'Adam',

  # schedules
  'constant',
  'exponential_decay',
  'inverse_time_decay',
  'polynomial_decay',
  'piecewise_constant',
]


class Optimizer(Base):
  """Base Optimizer Class.

  """
  target_backend = 'jax'

  def __init__(self, train_vars, lr, name):
    super(Optimizer, self).__init__(name=name)

    if isinstance(train_vars, ArrayCollector):
      train_vars = train_vars.subset(bm.TrainVar).unique()
    elif isinstance(train_vars, (list, tuple)):
      train_vars = ArrayCollector((f'_unknown{i}', var) for i, var in enumerate(train_vars))
      train_vars = train_vars.unique()
    else:
      raise ValueError
    self.lr = _make_schedule(lr)
    self.step = bm.Variable(bm.array([0]))
    self._train_vars = train_vars
    self.dynamic_vars = ArrayCollector(train_vars)  # dynamic variables

  def register_dynamical_vars(self, vars: dict):
    if not hasattr(self, 'dynamic_vars'):
      raise ValueError('Please super initialize the Optimizer first, then call "register_dynamical_vars()".')
    for key, var in vars.items():
      if key in self.dynamic_vars:
        if id(self.dynamic_vars[key]) != id(var):
          raise ValueError(f'Name "{key}" conflicts: same name for {var} and {self.dynamic_vars[key]}.')
      self.dynamic_vars[key] = var

  def vars(self, method='absolute'):
    gather = ArrayCollector(self.dynamic_vars)
    gather.update(super(Optimizer, self).vars(method=method))
    return gather

  def update(self, grads):
    if len(grads) != len(self._train_vars):
      raise ValueError('Expecting as many gradients as trainable variables')
    self.step += 1
    lr = self.lr(self.step[0])
    return lr


class SGD(Optimizer):
  """Stochastic gradient descent optimizer.
  """

  def __init__(self, lr, train_vars, name=None):
    super(SGD, self).__init__(lr=lr, train_vars=train_vars, name=name)

  def update(self, grads: dict, **kwargs):
    lr = super(SGD, self).update(grads)
    for key, p in self._train_vars.items():
      p.value -= lr * grads[key]


class Momentum(Optimizer):
  """Momentum optimizer.

  """

  def __init__(self, lr, train_vars, momentum, name=None):
    super(Momentum, self).__init__(lr=lr, train_vars=train_vars, name=name)

    self.momentum = momentum
    ms = dict((key + '_m', bm.Variable(bm.zeros_like(x))) for key, x in self._train_vars.items())
    self.register_dynamical_vars(ms)

  def update(self, grads: dict, **kwargs):
    lr = super(Momentum, self).update(grads)
    for key, p in self._train_vars.items():
      m = self.dynamic_vars[key + '_m']
      g = grads[key]
      m.value = g + self.momentum * m.value
      p.value -= lr * m.value


class NesterovMomentum(Optimizer):
  def __init__(self, lr, train_vars, momentum, name=None):
    super(NesterovMomentum, self).__init__(lr=lr, train_vars=train_vars, name=name)

    self.momentum = momentum
    ms = dict((key + '_m', bm.Variable(bm.zeros_like(x))) for key, x in self._train_vars.items())
    self.register_dynamical_vars(ms)

  def update(self, grads: dict, **kwargs):
    lr = super(NesterovMomentum, self).update(grads)
    for key, p in self._train_vars.items():
      m = self.dynamic_vars[key + '_m']
      g = grads[key]
      m.value = g + self.momentum * m.value
      p.value -= lr * (g + self.momentum * m.value)


class Adam(Optimizer):
  """Adam optimizer.

  Adam [1]_ - a stochastic gradient descent method (SGD) that computes
  individual adaptive learning rates for different parameters from estimates of
  first- and second-order moments of the gradients.

  Parameters
  ----------
  beta1: optional, float
    A positive scalar value for beta_1, the exponential decay rate
    for the first moment estimates (default 0.9).
  beta2: optional, float
    A positive scalar value for beta_2, the exponential decay rate
    for the second moment estimates (default 0.999).
  eps: optional, float
    A positive scalar value for epsilon, a small constant for
    numerical stability (default 1e-8).
  name : optional, str
    The optimizer name.

  References
  ----------
  .. [1] Kingma, D. P., & Ba, J. (2014). Adam: A method for stochastic optimization. arXiv preprint arXiv:1412.6980.
  """

  def __init__(self, lr, train_vars, beta1=0.9, beta2=0.999, eps=1e-8, name=None):
    super(Adam, self).__init__(lr=lr, train_vars=train_vars, name=name)

    self.beta1 = beta1
    self.beta2 = beta2
    self.eps = eps
    ms = dict((key + '_m', bm.Variable(bm.zeros_like(x))) for key, x in self._train_vars.items())
    vs = dict((key + '_v', bm.Variable(bm.zeros_like(x))) for key, x in self._train_vars.items())
    self.register_dynamical_vars(ms)
    self.register_dynamical_vars(vs)

  def update(self, grads: dict, **kwargs):
    lr = super(Adam, self).update(grads)
    lr *= jn.sqrt(1 - self.beta2 ** self.step[0]) / (1 - self.beta1 ** self.step[0])
    for key, p in self._train_vars.items():
      m = self.dynamic_vars[key + '_m']
      v = self.dynamic_vars[key + '_v']
      g = grads[key]
      m.value = self.beta1 * m.value + (1 - self.beta1) * g
      v.value = self.beta2 * v.value + (1 - self.beta2) * g ** 2
      p.value -= lr * m.value / jn.sqrt(v.value + self.eps)


# learning rate schedules

def constant(lr):
  def schedule(i):
    return lr

  return schedule


def exponential_decay(lr, decay_steps, decay_rate):
  def schedule(i):
    return lr * decay_rate ** (i / decay_steps)

  return schedule


def inverse_time_decay(lr, decay_steps, decay_rate, staircase=False):
  if staircase:
    def schedule(i):
      return lr / (1 + decay_rate * jn.floor(i / decay_steps))
  else:
    def schedule(i):
      return lr / (1 + decay_rate * i / decay_steps)
  return schedule


def polynomial_decay(lr, decay_steps, final_lr, power=1.0):
  def schedule(i):
    i = jn.minimum(i, decay_steps)
    step_mult = (1 - i / decay_steps) ** power
    return step_mult * (lr - final_lr) + final_lr

  return schedule


def piecewise_constant(boundaries, values):
  boundaries = jn.array(boundaries)
  values = jn.array(values)
  if not boundaries.ndim == values.ndim == 1:
    raise ValueError("boundaries and values must be sequences")
  if not boundaries.shape[0] == values.shape[0] - 1:
    raise ValueError("boundaries length must be one shorter than values length")

  def schedule(i):
    return values[jn.sum(i > boundaries)]

  return schedule


def _make_schedule(scalar_or_schedule):
  if callable(scalar_or_schedule):
    return scalar_or_schedule
  elif isinstance(scalar_or_schedule, (int, float)):
    return constant(scalar_or_schedule)
  else:
    raise TypeError(type(scalar_or_schedule))



# -*- coding: utf-8 -*-


from brainpy import math
from brainpy.simulation.brainobjects.neuron import NeuGroup
from brainpy.simulation.initialize import Initializer, XavierNormal, Uniform, ZeroInit

__all__ = [
  'LinearReadout'
]


class LinearReadout(NeuGroup):
  """Neuron group to readout information linearly.

  Parameters
  ----------
  num_hidden : int
    The neuron group size.
  num_input : int
    The input size.
  w_init : Initializer
    Initializer for the weights.
  b_init : Initializer
    Initializer for the bias.
  has_bias : bool
    Whether has the bias to compute.
  s_init : Initializer
    Initializer for variable states.
  train_mask : optional, math.ndarray
    The training mask for the weights.
  """

  def __init__(self, num_hidden, num_input, num_batch=1,
               w_init=XavierNormal(), b_init=ZeroInit(),
               has_bias=True, s_init=Uniform(), train_mask=None, **kwargs):
    super(LinearReadout, self).__init__(size=num_hidden, **kwargs)

    # parameters
    self.w_init = w_init
    self.b_init = b_init
    self.s_init = s_init
    self.num_input = num_input
    self.has_bias = has_bias

    # weights
    self.w = math.TrainVar(w_init((num_input, self.num)))
    if has_bias: self.b = math.TrainVar(b_init((self.num,)))

    if train_mask is not None:
      assert train_mask.shape == self.w.shape
      self.train_mask = train_mask


    # variables
    self.s = math.Variable(self.s_init((num_batch, self.num)))

  def update(self, x, **kwargs):
    if self.has_bias: self.s[:] = x @ self.w + self.b
    else: self.s[:] = x @ self.w
    return self.s



# -*- coding: utf-8 -*-

"""
The compilation tools for JAX backend.

1. Just-In-Time compilation is implemented by the 'jit()' function
2. Vectorize compilation is implemented by the 'vmap()' function
3. Parallel compilation is implemented by the 'pmap()' function

"""

import functools
import logging

import jax
import jax.numpy as jnp
import numpy as np
from jax.interpreters.partial_eval import DynamicJaxprTracer
from jax.interpreters.partial_eval import JaxprTracer
from jax.interpreters.pxla import ShardedDeviceArray

from brainpy import errors
from brainpy.base.base import Base
from brainpy.math.jax.random import RandomState
from brainpy.base.collector import ArrayCollector
from brainpy.tools.codes import change_func_name

__all__ = [
  'jit',
  'vmap',
  'pmap',
]

logger = logging.getLogger('brainpy.math.jax.compilation')


def _make_jit(func, vars, static_argnums, static_argnames=None, device=None,
              backend=None, donate_argnums=(), inline=False, f_name=None):
  @functools.partial(jax.jit,
                     static_argnums=static_argnums,
                     static_argnames=static_argnames,
                     device=device, backend=backend,
                     donate_argnums=donate_argnums,
                     inline=inline)
  def jitted_func(variable_data, *args, **kwargs):
    vars.assign(variable_data)
    out = func(*args, **kwargs)
    changes = vars.dict()
    return out, changes

  def call(*args, **kwargs):
    variable_data = vars.dict()
    out, changes = jitted_func(variable_data, *args, **kwargs)
    vars.assign(changes)
    return out

  return change_func_name(name=f_name, f=call) if f_name else call


def jit(obj_or_func, vars=None, static_argnums=None, static_argnames=None, device=None,
        backend=None, donate_argnums=(), inline=False, **kwargs):
  """JIT (Just-In-Time) Compilation for JAX backend.

  This function has the same ability to Just-In-Time compile a pure function,
  but it can also JIT compile a :py:class:`brainpy.DynamicalSystem`, or a
  :py:class:`brainpy.Base` object, or a bounded method of a
  :py:class:`brainpy.Base` object.

  If you are using "numpy", please refer to the JIT compilation
  in NumPy backend `bp.math.numpy.jit() <brainpy.math.numpy.jit.rst>`_.

  Examples
  --------

  You can JIT a :py:class:`brainpy.DynamicalSystem`

  >>> import brainpy as bp
  >>>
  >>> class LIF(bp.NeuGroup):
  >>>   pass
  >>> lif = bp.math.jit(LIF(10))

  You can JIT a :py:class:`brainpy.Base` object with ``__call__()`` implementation.

  >>> mlp = bp.nets.MLP((10, 100, 10))
  >>> jit_mlp = bp.math.jit(mlp)

  You can also JIT a bounded method of a :py:class:`brainpy.Base` object.

  >>> class Hello(bp.Base):
  >>>   def __init__(self):
  >>>     super(Hello, self).__init__()
  >>>     self.a = bp.math.Variable(bp.math.array(10.))
  >>>     self.b = bp.math.Variable(bp.math.array(2.))
  >>>   def transform(self):
  >>>     return self.a ** self.b
  >>>
  >>> test = Hello()
  >>> bp.math.jit(test.transform)

  Further, you can JIT a normal function, just used like in JAX.

  >>> @bp.math.jit
  >>> def selu(x, alpha=1.67, lmbda=1.05):
  >>>   return lmbda * bp.math.where(x > 0, x, alpha * bp.math.exp(x) - alpha)


  Notes
  -----

  There are several notes:

  1. Avoid using scalar in Variable, TrainVar, etc.


  Parameters
  ----------
  obj_or_func : Base, function
    The instance of Base or a function.
  vars : optional, dict
    These variables will be changed in the function, or needed in the computation.
  static_argnums : optional, int, list, tuple, dict
    An optional int or collection of ints that specify which positional arguments to treat
    as static (compile-time constant).
  static_argnames : optional, str, list, tuple, dict
    An optional string or collection of strings specifying which named arguments to treat
    as static (compile-time constant). See the comment on ``static_argnums`` for details.
    If not provided but ``static_argnums`` is set, the default is based on calling
    ``inspect.signature(fun)`` to find corresponding named arguments.
  device: optional, Any
    This is an experimental feature and the API is likely to change.
    Optional, the Device the jitted function will run on. (Available devices
    can be retrieved via :py:func:`jax.devices`.) The default is inherited
    from XLA's DeviceAssignment logic and is usually to use
    ``jax.devices()[0]``.
  backend: optional, str, dict
    This is an experimental feature and the API is likely to change. Optional,
    a string representing the XLA backend: ``'cpu'``, ``'gpu'``, or ``'tpu'``.
  donate_argnums: optional, int, dict, tuple, list
    Specify which arguments are "donated" to the computation.
    It is safe to donate arguments if you no longer need them once the
    computation has finished. In some cases XLA can make use of donated
    buffers to reduce the amount of memory needed to perform a computation,
    for example recycling one of your input buffers to store a result. You
    should not reuse buffers that you donate to a computation, JAX will raise
    an error if you try to. By default, no arguments are donated.
  inline: bool
    Specify whether this function should be inlined into enclosing
    jaxprs (rather than being represented as an application of the xla_call
    primitive with its own subjaxpr). Default False.

  Returns
  -------
  ds_of_func : Base, function
    A wrapped version of Base object or function, set up for just-in-time compilation.
  """
  from brainpy.simulation.brainobjects.base import DynamicalSystem

  if isinstance(obj_or_func, DynamicalSystem):
    if len(obj_or_func.steps):  # DynamicalSystem has step functions

      # dynamical variables
      vars = (vars or obj_or_func.vars().unique())

      # static arguments by num
      if static_argnums is None:
        static_argnums = {key: () for key in obj_or_func.steps.keys()}
      elif isinstance(static_argnums, int):
        static_argnums = {key: (static_argnums + 1,) for key in obj_or_func.steps.keys()}
      elif isinstance(static_argnums, (tuple, list)) and isinstance(static_argnums[0], int):
        static_argnums = {key: tuple(x + 1 for x in static_argnums) for key in obj_or_func.steps.keys()}
      assert isinstance(static_argnums, dict)

      # static arguments by name
      if static_argnames is None:
        static_argnames = {key: None for key in obj_or_func.steps.keys()}
      elif isinstance(static_argnames, str):
        static_argnames = {key: (static_argnames,) for key in obj_or_func.steps.keys()}
      elif isinstance(static_argnames, (tuple, list)) and isinstance(static_argnames[0], str):
        static_argnames = {key: static_argnames for key in obj_or_func.steps.keys()}
      assert isinstance(static_argnames, dict)

      # donate arguments by num
      if donate_argnums is None:
        donate_argnums = {key: () for key in obj_or_func.steps.keys()}
      elif isinstance(donate_argnums, int):
        donate_argnums = {key: (donate_argnums + 1,) for key in obj_or_func.steps.keys()}
      elif isinstance(donate_argnums, (tuple, list)):
        donate_argnums = {key: tuple(x + 1 for x in donate_argnums) for key in obj_or_func.steps.keys()}
      assert isinstance(donate_argnums, dict)

      # inline
      if not isinstance(inline, dict):
        inline = {key: inline for key in obj_or_func.steps.keys()}
      assert isinstance(inline, dict)

      # jit functions
      for key in list(obj_or_func.steps.keys()):
        jitted_func = _make_jit(vars=vars,
                                func=obj_or_func.steps[key],
                                static_argnums=static_argnums[key],
                                static_argnames=static_argnames[key],
                                device=device,
                                backend=backend,
                                donate_argnums=donate_argnums[key],
                                inline=inline[key],
                                f_name=key)
        obj_or_func.steps.replace(key, jitted_func)
      return obj_or_func

  if callable(obj_or_func):
    if vars is not None:
      vars = vars
    elif isinstance(obj_or_func, Base):
      vars = obj_or_func.vars().unique()
    elif hasattr(obj_or_func, '__self__') and isinstance(obj_or_func.__self__, Base):
      vars = obj_or_func.__self__.vars().unique()
    else:
      vars = ArrayCollector()

    if len(vars) == 0:  # pure function
      return jax.jit(obj_or_func,
                     static_argnums=static_argnums,
                     static_argnames=static_argnames,
                     device=device,
                     backend=backend,
                     donate_argnums=donate_argnums,
                     inline=inline)

    else:  # Base object which implements __call__, or bounded method of Base object

      static_argnums = tuple(x + 1 for x in sorted(static_argnums or ()))
      return _make_jit(vars=vars,
                       func=obj_or_func,
                       static_argnums=static_argnums,
                       static_argnames=static_argnames,
                       device=device,
                       backend=backend,
                       donate_argnums=donate_argnums,
                       inline=inline)

  else:
    raise errors.BrainPyError(f'Only support instance of {Base.__name__}, or a callable '
                              f'function, but we got {type(obj_or_func)}.')


def _make_vmap(func, dyn_vars, rand_vars, in_axes, out_axes,
               batch_idx, axis_name, reduce_func, f_name=None):
  @functools.partial(jax.vmap, in_axes=in_axes, out_axes=out_axes, axis_name=axis_name)
  def vmapped_func(dyn_data, rand_data, *args, **kwargs):
    dyn_vars.assign(dyn_data)
    rand_vars.assign(rand_data)
    out = func(*args, **kwargs)
    dyn_changes = dyn_vars.dict()
    rand_changes = rand_vars.dict()
    return out, dyn_changes, rand_changes

  def call(*args, **kwargs):
    dyn_data = dyn_vars.dict()
    n = args[batch_idx[0]].shape[batch_idx[1]]
    rand_data = {key: val.split_keys(n) for key, val in rand_vars.items()}
    out, dyn_changes, rand_changes = vmapped_func(dyn_data, rand_data, *args, **kwargs)
    for key, v in dyn_changes.items():
      dyn_vars[key] = reduce_func(v)
    for key, v in rand_changes.items():
      rand_vars[key] = reduce_func(v)
    return out

  return change_func_name(name=f_name, f=call) if f_name else call


def vmap(obj_or_func, vars=None, vars_batched=None,
         in_axes=0, out_axes=0, axis_name=None, reduce_func=None):
  """Vectorization compilation in JAX backend.

  Vectorized compile a function or a module to run in parallel on a single device.

  Examples
  --------

  Parameters
  ----------
  obj_or_func : Base, function
    The function or the module to compile.
  vars_needed : dict
  vars_batched : dict
  in_axes : optional, int, tuple/list/dict
    Specify which input array axes to map over. If each positional argument to
    ``obj_or_func`` is an array, then ``in_axes`` can be an integer, a None,
    or a tuple of integers and Nones with length equal to the number of
    positional arguments to ``obj_or_func``. An integer or ``None``
    indicates which array axis to map over for all arguments (with ``None``
    indicating not to map any axis), and a tuple indicates which axis to map
    for each corresponding positional argument. Axis integers must be in the
    range ``[-ndim, ndim)`` for each array, where ``ndim`` is the number of
    dimensions (axes) of the corresponding input array.

    If the positional arguments to ``obj_or_func`` are container types, the
    corresponding element of ``in_axes`` can itself be a matching container,
    so that distinct array axes can be mapped for different container
    elements. ``in_axes`` must be a container tree prefix of the positional
    argument tuple passed to ``obj_or_func``.

    At least one positional argument must have ``in_axes`` not None. The sizes
    of the mapped input axes for all mapped positional arguments must all be
    equal.

    Arguments passed as keywords are always mapped over their leading axis
    (i.e. axis index 0).
  out_axes : optional, int, tuple/list/dict
    Indicate where the mapped axis should appear in the output. All outputs
    with a mapped axis must have a non-None ``out_axes`` specification. Axis
    integers must be in the range ``[-ndim, ndim)`` for each output array,
    where ``ndim`` is the number of dimensions (axes) of the array returned
    by the :func:`vmap`-ed function, which is one more than the number of
    dimensions (axes) of the corresponding array returned by ``obj_or_func``.
  axis_name : optional

  Returns
  -------
  obj_or_func : Base, function
    Batched/vectorized version of ``obj_or_func`` with arguments that correspond to
    those of ``obj_or_func``, but with extra array axes at positions indicated by
    ``in_axes``, and a return value that corresponds to that of ``obj_or_func``, but
    with extra array axes at positions indicated by ``out_axes``.

  """
  from brainpy.simulation.brainobjects.base import DynamicalSystem

  if isinstance(obj_or_func, DynamicalSystem):
    if len(obj_or_func.steps):  # DynamicalSystem has step functions

      # dynamical variables
      vars = (vars or obj_or_func.vars().unique())
      dyn_vars, rand_vars = ArrayCollector(), ArrayCollector()
      for key, val in vars.items():
        if isinstance(val, RandomState):
          rand_vars[key] = val
        else:
          dyn_vars[key] = val

      # in axes
      if in_axes is None:
        in_axes = {key: (None, 0) for key in obj_or_func.steps.keys()}
      elif isinstance(in_axes, int):
        in_axes = {key: (None, 0, in_axes) for key in obj_or_func.steps.keys()}
      elif isinstance(in_axes, (tuple, list)):
        in_axes = {key: (None, 0) + tuple(in_axes) for key in obj_or_func.steps.keys()}
      elif isinstance(in_axes, dict):
        keys = list(obj_or_func.steps.keys())
        if keys[0] not in in_axes:
          in_axes = {key: (None, 0, in_axes) for key in keys}
        else:
          in_axes = {key: (None, 0) + tuple(in_axes[key]) for key in keys}
      assert isinstance(in_axes, dict)

      # batch size index
      batch_idx = {}
      for key, axes in in_axes.items():
        for i, axis in enumerate(axes[2:]):
          if axis is not None:
            batch_idx[key] = (i, axis)
            break
        else:
          raise ValueError(f'Found no batch axis: {axes}.')

      # out axes
      if out_axes is None:
        out_axes = {key: 0 for key in obj_or_func.steps.keys()}
      elif isinstance(out_axes, int):
        out_axes = {key: out_axes for key in obj_or_func.steps.keys()}
      elif isinstance(out_axes, (tuple, list)):
        out_axes = {key: tuple(out_axes) + (0, 0) for key in obj_or_func.steps.keys()}
      elif isinstance(out_axes, dict):
        keys = list(obj_or_func.steps.keys())
        if keys[0] not in out_axes:
          out_axes = {key: (out_axes, 0, 0) for key in keys}
        else:
          out_axes = {key: tuple(out_axes[key]) + (0, 0) for key in keys}
      assert isinstance(out_axes, dict)

      # reduce_func
      if reduce_func is None:
        reduce_func = lambda x: x.mean(axis=0)

      # vectorized map functions
      for key in obj_or_func.steps.keys():
        obj_or_func.steps[key] = _make_vmap(func=obj_or_func.steps[key],
                                            dyn_vars=dyn_vars,
                                            rand_vars=rand_vars,
                                            in_axes=in_axes[key],
                                            out_axes=out_axes[key],
                                            axis_name=axis_name,
                                            batch_idx=batch_idx[key],
                                            reduce_func=reduce_func,
                                            f_name=key)

      return obj_or_func

  if callable(obj_or_func):
    if vars is not None:
      vars = vars
    elif isinstance(obj_or_func, Base):  # Base has '__call__()' implementation
      vars = obj_or_func.vars().unique()
    elif hasattr(obj_or_func, '__self__'):
      if isinstance(obj_or_func.__self__, Base):
        vars = obj_or_func.__self__.vars().unique()

    if vars is None:
      return jax.vmap(obj_or_func,
                      in_axes=in_axes,
                      out_axes=out_axes,
                      axis_name=axis_name)

    else:
      # dynamical variables
      dyn_vars, rand_vars = ArrayCollector(), ArrayCollector()
      for key, val in vars.items():
        if isinstance(val, RandomState):
          rand_vars[key] = val
        else:
          dyn_vars[key] = val

      # in axes
      if in_axes is None:
        in_axes = (None, 0)
      elif isinstance(in_axes, (int, dict)):
        in_axes = (None, 0, in_axes)
      elif isinstance(in_axes, (tuple, list)):
        in_axes = (None, 0) + tuple(in_axes)
      assert isinstance(in_axes, (tuple, list))

      # batch size index
      batch_idx = {}
      for key, axes in batch_idx.items():
        for i, axis in enumerate(axes[2:]):
          if axis is not None:
            batch_idx[key] = (i, axis)
            break
        else:
          raise ValueError(f'Found no batch axis: {axes}.')

      # out axes
      if out_axes is None:
        out_axes = 0
      elif isinstance(out_axes, (int, dict)):
        out_axes = (out_axes, 0, 0)
      elif isinstance(out_axes, (tuple, list)):
        out_axes = tuple(out_axes) + (0, 0)
      assert isinstance(out_axes, (list, tuple))

      # reduce_func
      if reduce_func is None:
        reduce_func = lambda x: x.mean(axis=0)

      # jit function
      return _make_vmap(func=obj_or_func,
                        dyn_vars=dyn_vars,
                        rand_vars=rand_vars,
                        in_axes=in_axes,
                        out_axes=out_axes,
                        axis_name=axis_name,
                        batch_idx=batch_idx,
                        reduce_func=reduce_func)

  else:
    raise errors.BrainPyError(f'Only support instance of {Base.__name__}, or a callable '
                              f'function, but we got {type(obj_or_func)}.')


def _device_reshape(x):
  """Reshape an input array in order to broadcast to multiple devices."""
  num_device = jax.local_device_count()

  if not hasattr(x, 'ndim'):
    raise errors.BrainPyError(f'Expected JaxArray, got {type(x)}. If you are trying to pass a scalar to '
                              f'parallel, first convert it to a JaxArray, for example np.float(0.5)')
  if x.ndim == 0:
    return np.broadcast_to(x, [num_device])
  if x.shape[0] % num_device != 0:
    raise errors.BrainPyError(f'Must be able to equally divide batch {x.shape} among '
                              f'{num_device} devices, but does not go equally.')
  return x.reshape((num_device, x.shape[0] // num_device) + x.shape[1:])


def _make_pmap(func, dyn_vars, rand_vars, reduce_func, axis_name=None, in_axes=0,
               out_axes=0, static_broadcasted_argnums=(), devices=None, backend=None,
               axis_size=None, donate_argnums=(), global_arg_shapes=None, f_name=None):
  @functools.partial(jax.pmap, in_axes=in_axes, out_axes=out_axes, axis_name=axis_name,
                     static_broadcasted_argnums=static_broadcasted_argnums, devices=devices,
                     backend=backend, axis_size=axis_size, donate_argnums=donate_argnums,
                     global_arg_shapes=global_arg_shapes)
  def pmapped_func(dyn_data, rand_data, *args, **kwargs):
    dyn_vars.assign(dyn_data)
    rand_vars.assign(rand_data)
    out = func(*args, **kwargs)
    dyn_changes = dyn_vars.dict()
    rand_changes = rand_vars.dict()
    return out, dyn_changes, rand_changes

  def call(*args):
    un_replicated = [k for k, v in dyn_vars.items()
                     if not isinstance(v.value, (ShardedDeviceArray, JaxprTracer, DynamicJaxprTracer))]
    if len(un_replicated):
      raise errors.BrainPyError(f'Some variables were not replicated: {un_replicated}.'
                                f'did you forget to call xx.replicate() on them?')
    _args = []
    for i, x in enumerate(args):
      if i + 2 in static_broadcasted_argnums:
        _args.append(x)
      else:
        _args.append(jax.tree_map(_device_reshape, [x])[0])
    dyn_data = dyn_vars.dict()
    rand_data = rand_vars.dict()
    output, dyn_changes, rand_changes = pmapped_func(dyn_data, rand_data, *_args)
    dyn_vars.assign(dyn_changes)
    rand_vars.assign(rand_changes)
    return jax.tree_map(reduce_func, output)

  return change_func_name(name=f_name, f=call) if f_name else call


def pmap(obj_or_func, vars=None, axis_name=None, in_axes=0, out_axes=0, static_broadcasted_argnums=(),
         devices=None, backend=None, axis_size=None, donate_argnums=(), global_arg_shapes=None,
         reduce_func=None):
  """Parallel compilation in JAX backend.

  Parallel compile a function or a module to run on multiple devices in parallel.

  Parameters
  ----------
  obj_or_func
  axis_name
  in_axes
  out_axes
  static_broadcasted_argnums
  devices
  backend
  axis_size
  donate_argnums
  global_arg_shapes

  Returns
  -------


  Examples
  --------


  """
  from brainpy.simulation.brainobjects.base import DynamicalSystem

  if isinstance(obj_or_func, DynamicalSystem):
    if len(obj_or_func.steps):  # DynamicalSystem has step functions

      # dynamical variables
      all_vars = (vars or obj_or_func.vars().unique())
      dyn_vars = ArrayCollector()
      rand_vars = ArrayCollector()
      for key, val in all_vars.items():
        if isinstance(val, RandomState):
          rand_vars[key] = val
        else:
          dyn_vars[key] = val

      # reduce function
      if reduce_func is None:
        reduce_func = jnp.concatenate

      # static broadcast-ed arguments
      if static_broadcasted_argnums is None:
        static_broadcasted_argnums = ()
      elif isinstance(static_broadcasted_argnums, int):
        static_broadcasted_argnums = (static_broadcasted_argnums + 2,)
      elif isinstance(static_broadcasted_argnums, (tuple, list)):
        static_broadcasted_argnums = tuple(argnum + 2 for argnum in static_broadcasted_argnums)
      assert isinstance(static_broadcasted_argnums, (tuple, list))

      # jit functions
      for key in obj_or_func.steps.keys():
        step = obj_or_func.steps[key]
        obj_or_func.steps[key] = _make_pmap(dyn_vars=dyn_vars,
                                            rand_vars=rand_vars,
                                            func=step,
                                            axis_name=axis_name,
                                            in_axes=in_axes,
                                            out_axes=out_axes,
                                            static_broadcasted_argnums=static_broadcasted_argnums,
                                            devices=devices,
                                            backend=backend,
                                            axis_size=axis_size,
                                            donate_argnums=donate_argnums,
                                            global_arg_shapes=global_arg_shapes,
                                            reduce_func=reduce_func,
                                            f_name=key)
      return obj_or_func

  if callable(obj_or_func):
    if vars is not None:
      vars = vars
    elif isinstance(obj_or_func, Base):  # Base has '__call__()' implementation
      vars = obj_or_func.vars().unique()
    elif hasattr(obj_or_func, '__self__'):
      if isinstance(obj_or_func.__self__, Base):
        vars = obj_or_func.__self__.vars().unique()

    if vars is None:
      return jax.pmap(obj_or_func,
                      axis_name=axis_name,
                      in_axes=in_axes,
                      out_axes=out_axes,
                      static_broadcasted_argnums=static_broadcasted_argnums,
                      devices=devices,
                      backend=backend,
                      axis_size=axis_size,
                      donate_argnums=donate_argnums,
                      global_arg_shapes=global_arg_shapes)
    else:
      # dynamical variables
      dyn_vars = ArrayCollector()
      rand_vars = ArrayCollector()
      for key, val in vars.items():
        if isinstance(val, RandomState):
          rand_vars[key] = val
        else:
          dyn_vars[key] = val

      # static broadcast-ed arguments
      if static_broadcasted_argnums is None:
        static_broadcasted_argnums = ()
      elif isinstance(static_broadcasted_argnums, int):
        static_broadcasted_argnums = (static_broadcasted_argnums + 2,)
      elif isinstance(static_broadcasted_argnums, (tuple, list)):
        static_broadcasted_argnums = tuple(argnum + 2 for argnum in static_broadcasted_argnums)
      assert isinstance(static_broadcasted_argnums, (tuple, list))

      # reduce function
      if reduce_func is None:
        reduce_func = jnp.concatenate

      # jit function
      obj_or_func.__call__ = _make_pmap(dyn_vars=dyn_vars,
                                        rand_vars=rand_vars,
                                        func=obj_or_func,
                                        axis_name=axis_name,
                                        in_axes=in_axes,
                                        out_axes=out_axes,
                                        static_broadcasted_argnums=static_broadcasted_argnums,
                                        devices=devices,
                                        backend=backend,
                                        axis_size=axis_size,
                                        donate_argnums=donate_argnums,
                                        global_arg_shapes=global_arg_shapes,
                                        reduce_func=reduce_func)
      return obj_or_func

  else:
    raise errors.BrainPyError(f'Only support instance of {Base.__name__}, or a callable function, '
                              f'but we got {type(obj_or_func)}.')



# -*- coding: utf-8 -*-


from brainpy import math
from brainpy.simulation.brainobjects.base import DynamicalSystem
from brainpy.simulation.initialize import XavierNormal, Initializer, ZeroInit

__all__ = [
  'Dense'
]


class Dense(DynamicalSystem):
  """A fully connected layer implemented as the dot product of inputs and weights.

  Parameters
  ----------
  num_hidden : int
    The neuron group size.
  num_input : int
    The input size.
  w_init : Initializer
    Initializer for the weights.
  b_init : Initializer
    Initializer for the bias.
  has_bias : bool
    Whether has the bias to compute.
  steps : tuple of str, tuple of function, dict of (str, function), optional
      The callable function, or a list of callable functions.
  monitors : None, list, tuple, datastructures.Monitor
      Variables to monitor.
  name : str, optional
      The name of the dynamic system.
  """

  def __init__(self, num_hidden, num_input, w_init=XavierNormal(),
               b_init=ZeroInit(), has_bias=True, **kwargs):
    super(Dense, self).__init__(**kwargs)

    # parameters
    self.has_bias = has_bias
    self.num_input = num_input
    self.num_hidden = num_hidden

    # variables
    self.w = math.TrainVar(w_init((num_input, num_hidden)))
    if has_bias: self.b = math.TrainVar(b_init((num_hidden,)))

  def update(self, x, **kwargs):
    """Returns the results of applying the linear transformation to input x."""
    if self.has_bias: return x @ self.w + self.b
    else: return x @ self.w



# -*- coding: utf-8 -*-

import numpy as np

from brainpy import math
from .base import Initializer

__all__ = [
  'Normal',
  'Uniform',
  'Orthogonal',
  'KaimingNormal',
  'KaimingNormalTruncated',
  'XavierNormal',
  'XavierNormalTruncated',
  'TruncatedNormal',
]


class Normal(Initializer):
  """Initialize weights with normal distribution.

  Parameters
  ----------
  gain : float
    The gain of the derivation of the normal distribution.

  """
  def __init__(self, gain=1., dtype=None):
    self.gain = gain
    super(Normal, self).__init__(dtype=dtype)

  def __call__(self, shape):
    weights = math.random.normal(size=shape, scale=self.gain * np.sqrt(1 / np.prod(shape)))
    return math.asarray(weights, dtype=self.dtype)


class Uniform(Initializer):
  """Initialize weights with uniform distribution.

  Parameters
  ----------
  min_val : float
    The lower limit of the uniform distribution.
  max_val : float
    The upper limit of the uniform distribution.

  """
  def __init__(self, min_val=0., max_val=1., dtype=None):
    super(Uniform, self).__init__(dtype=dtype)
    self.min_val = min_val
    self.max_val = max_val

  def __call__(self, shape):
    r = math.random.uniform(low=self.min_val, high=self.max_val, size=shape)
    return math.asarray(r, dtype=self.dtype)


class Orthogonal(Initializer):
  """Returns a uniformly distributed orthogonal tensor from
    `Exact solutions to the nonlinear dynamics of learning in deep linear neural networks
    <https://openreview.net/forum?id=_wzZwKpTDF_9C>`_.

    Args:
        shape: shape of the output tensor.
        gain: optional scaling factor.
        axis: the orthogonalizarion axis

    Returns:
        An orthogonally initialized tensor.
        These tensors will be row-orthonormal along the access specified by
        ``axis``. If the rank of the weight is greater than 2, the shape will be
        flattened in all other dimensions and then will be row-orthonormal along the
        final dimension. Note that this only works if the ``axis`` dimension is
        larger, otherwise the tensor will be transposed (equivalently, it will be
        column orthonormal instead of row orthonormal).
        If the shape is not square, the matrices will have orthonormal rows or
        columns depending on which side is smaller.
    """

  def __init__(self, gain=1., axis=-1, dtype=None):
    super(Orthogonal, self).__init__(dtype=dtype)
    self.gain = gain
    self.axis = axis

  def __call__(self, shape):
    n_rows = shape[self.axis]
    n_cols = np.prod(shape) // n_rows
    matrix_shape = (n_rows, n_cols) if n_rows > n_cols else (n_cols, n_rows)
    norm_dst = math.random.normal(size=matrix_shape)
    q_mat, r_mat = np.linalg.qr(norm_dst)
    # Enforce Q is uniformly distributed
    q_mat *= np.sign(np.diag(r_mat))
    if n_rows < n_cols: q_mat = q_mat.T
    q_mat = np.reshape(q_mat, (n_rows,) + tuple(np.delete(shape, self.axis)))
    q_mat = np.moveaxis(q_mat, 0, self.axis)
    return self.gain * math.asarray(q_mat, dtype=self.dtype)


class KaimingNormal(Initializer):
  """Returns a tensor with values assigned using Kaiming He normal initializer from
    `Delving Deep into Rectifiers: Surpassing Human-Level Performance on ImageNet Classification
    <https://arxiv.org/abs/1502.01852>`_.

    Args:
        shape: shape of the output tensor.
        gain: optional scaling factor.

    Returns:
        Tensor initialized with normal random variables with standard deviation (gain * kaiming_normal_gain).
    """

  def __init__(self, gain=1., dtype=None):
    self.gain = gain
    super(KaimingNormal, self).__init__(dtype=dtype)

  def __call__(self, shape):
    gain = np.sqrt(1 / np.prod(shape[:-1]))
    res = math.random.normal(size=shape, scale=self.gain * gain)
    return math.asarray(res, dtype=self.dtype)


class KaimingNormalTruncated(Initializer):
  """Returns a tensor with values assigned using Kaiming He truncated normal initializer from
    `Delving Deep into Rectifiers: Surpassing Human-Level Performance on ImageNet Classification
    <https://arxiv.org/abs/1502.01852>`_.

    Args:
        shape: shape of the output tensor.
        lower: lower truncation of the normal.
        upper: upper truncation of the normal.
        gain: optional scaling factor.

    Returns:
        Tensor initialized with truncated normal random variables with standard
        deviation (gain * kaiming_normal_gain) and support [lower, upper].
    """

  def __init__(self, lower=-2., upper=2., gain=1., dtype=None):
    self.lower = lower
    self.upper = upper
    self.gain = gain
    super(KaimingNormalTruncated, self).__init__(dtype)

  def __call__(self, shape):
    truncated_std = scipy.stats.truncnorm.std(a=self.lower,
                                              b=self.upper,
                                              loc=0.,
                                              scale=1.)
    stddev = self.gain * np.sqrt(1 / np.prod(shape[:-1])) / truncated_std
    res = math.random.truncated_normal(size=shape,
                                       scale=stddev,
                                       lower=self.lower,
                                       upper=self.upper)
    return math.asarray(res, dtype=self.dtype)


class XavierNormal(Initializer):
  """Returns a tensor with values assigned using Xavier Glorot normal initializer from
    `Understanding the difficulty of training deep feedforward neural networks
    <http://proceedings.mlr.press/v9/glorot10a/glorot10a.pdf>`_.

    Args:
        shape: shape of the output tensor.
        gain: optional scaling factor.

    Returns:
        Tensor initialized with normal random variables with standard deviation (gain * xavier_normal_gain).
    """

  def __init__(self, gain=1., dtype=None):
    self.gain = gain
    super(XavierNormal, self).__init__(dtype=dtype)

  def __call__(self, shape):
    fan_in, fan_out = np.prod(shape[:-1]), shape[-1]
    gain = np.sqrt(2 / (fan_in + fan_out))
    res = math.random.normal(size=shape, scale=self.gain * gain)
    return math.asarray(res, dtype=self.dtype)


class XavierNormalTruncated(Initializer):
  """Returns a tensor with values assigned using Xavier Glorot truncated normal initializer from
    `Understanding the difficulty of training deep feedforward neural networks
    <http://proceedings.mlr.press/v9/glorot10a/glorot10a.pdf>`_.

    Args:
        shape: shape of the output tensor.
        lower: lower truncation of the normal.
        upper: upper truncation of the normal.
        gain: optional scaling factor.

    Returns:
        Tensor initialized with truncated normal random variables with standard
        deviation (gain * xavier_normal_gain) and support [lower, upper].
    """

  def __init__(self, lower=-2., upper=2., gain=1., dtype=None):
    self.lower = lower
    self.upper = upper
    self.gain = gain
    super(XavierNormalTruncated, self).__init__(dtype=dtype)

  def __call__(self, shape):
    truncated_std = scipy.stats.truncnorm.std(a=self.lower, b=self.upper, loc=0., scale=1)
    fan_in, fan_out = np.prod(shape[:-1]), shape[-1]
    gain = np.sqrt(2 / (fan_in + fan_out))
    stddev = self.gain * gain / truncated_std
    res = math.random.truncated_normal(size=shape,
                                       scale=stddev,
                                       lower=self.lower,
                                       upper=self.upper)
    return math.asarray(res, dtype=self.dtype)


class TruncatedNormal(Initializer):
  """Returns a tensor with values assigned using truncated normal initialization.

    Args:
        shape: shape of the output tensor.
        lower: lower truncation of the normal.
        upper: upper truncation of the normal.
        stddev: expected standard deviation.

    Returns:
        Tensor initialized with truncated normal random variables with standard
        deviation stddev and support [lower, upper].
    """

  def __init__(self, lower=-2., upper=2., scale=1., dtype=None):
    self.lower = lower
    self.upper = upper
    self.scale = scale
    super(TruncatedNormal, self).__init__(dtype=dtype)

  def __call__(self, shape, dtype=None):
    truncated_std = scipy.stats.truncnorm.std(a=self.lower, b=self.upper, loc=0., scale=1)
    res = math.random.truncated_normal(size=shape,
                                       scale=self.scale / truncated_std,
                                       lower=self.lower,
                                       upper=self.upper)
    return math.asarray(res, dtype=self.dtype)



# -*- coding: utf-8 -*-


import pytest

import brainpy as bp
from brainpy.integrators.ode import adaptive_rk
from brainpy.integrators.ode import exponential
from brainpy.integrators.ode import explicit_rk

bp.backend.set('numpy')


def test_ode():
    def lorenz_f(x, y, z, t, sigma=10, beta=8 / 3, rho=28):
        dx = sigma * (y - x)
        dy = x * (rho - z) - y
        dz = x * y - beta * z
        return dx, dy, dz

    for method in explicit_rk.__all__ + exponential.__all__:
        for var_type in bp.SUPPORTED_VAR_TYPE:
            print(f'"{method}" method, "{var_type}" var type:')
            if method == 'exponential_euler' and var_type == bp.SYSTEM_VAR:
                with pytest.raises(bp.errors.IntegratorError):
                    bp.odeint(f=lorenz_f, show_code=True, method=method, var_type=var_type)
            else:
                bp.odeint(f=lorenz_f, show_code=True, method=method, var_type=var_type)
            print()

    for method in adaptive_rk.__all__:
        for var_type in bp.SUPPORTED_VAR_TYPE:
            for adaptive in [True, False]:
                print(f'"{method}" method, "{var_type}" var type, adaptive = {adaptive}:')
                bp.odeint(f=lorenz_f, show_code=True, method=method, var_type=var_type, adaptive=adaptive)
                print()


def test_sde():
    def lorenz_g(x, y, z, t, sigma=10, beta=8 / 3, rho=28, p=0.1):
        return p * x, p * y, p * z

    def lorenz_f(x, y, z, t, sigma=10, beta=8 / 3, rho=28, p=0.1):
        dx = sigma * (y - x)
        dy = x * (rho - z) - y
        dz = x * y - beta * z
        return dx, dy, dz

    # for method in euler_and_milstein.__all__:
    for method in ['euler', 'milstein']:
        for sde_type in bp.SUPPORTED_SDE_TYPE:
            if method == 'heun' and sde_type == bp.ITO_SDE:
                continue
            for var_type in bp.SUPPORTED_VAR_TYPE:
                for wiener_type in bp.SUPPORTED_WIENER_TYPE:
                    print(f'"{method}" method, "{sde_type}" sde type, '
                          f'"{var_type}" var type, "{wiener_type}" wiener type:')
                    bp.sdeint(f=lorenz_f, g=lorenz_g, show_code=True, method=method,
                              var_type=var_type,
                              sde_type=sde_type,
                              wiener_type=wiener_type)
                    print()

    for method in ['exponential_euler', 'srk1w1_scalar', 'srk2w1_scalar', 'KlPl_scalar']:
        for var_type in bp.SUPPORTED_VAR_TYPE:
            for wiener_type in bp.SUPPORTED_WIENER_TYPE:
                with pytest.raises(bp.errors.IntegratorError):
                    bp.sdeint(f=lorenz_f, g=lorenz_g, show_code=True, method=method,
                              var_type=var_type,
                              sde_type=bp.STRA_SDE,
                              wiener_type=wiener_type)

    # for method in ['exponential_euler', 'srk1w1_scalar', 'srk2w1_scalar', 'KlPl_scalar']:



# test_sde()



# -*- coding: utf-8 -*-


import inspect

from brainpy import errors
from brainpy.math import profile

__all__ = [
  'attr_replace',
  'get_num_indent',
  'get_func_body_code',
  'get_args',
]


def attr_replace(attr):
  return attr.replace('.', '_')


def get_num_indent(code_string, spaces_per_tab=4):
  """Get the indent of a patch of source code.

  Parameters
  ----------
  code_string : str
      The code string.
  spaces_per_tab : int
      The spaces per tab.

  Returns
  -------
  num_indent : int
      The number of the indent.
  """
  lines = code_string.split('\n')
  min_indent = 1000
  for line in lines:
    if line.strip() == '':
      continue
    line = line.replace('\t', ' ' * spaces_per_tab)
    num_indent = len(line) - len(line.lstrip())
    if num_indent < min_indent:
      min_indent = num_indent
  return min_indent


def get_func_body_code(code_string, lambda_func=False):
  """Get the main body code of a function.

  Parameters
  ----------
  code_string : str
      The code string of the function.
  lambda_func : bool
      Whether the code comes from a lambda function.

  Returns
  -------
  code_body : str
      The code body.
  """
  if lambda_func:
    splits = code_string.split(':')
    if len(splits) != 2:
      raise ValueError(f'Can not parse function: \n{code_string}')
    main_code = f'return {":".join(splits[1:])}'
  else:
    func_codes = code_string.split('\n')
    idx = 0
    for i, line in enumerate(func_codes):
      idx += 1
      line = line.replace(' ', '')
      if '):' in line:
        break
    else:
      raise ValueError(f'Can not parse function: \n{code_string}')
    main_code = '\n'.join(func_codes[idx:])
  return main_code


def get_args(f):
  """Get the function arguments.

  Parameters
  ----------
  f : callable
      The function.

  Returns
  -------
  args : tuple
      The variable names, the other arguments, and the original args.
  """

  # 1. get the function arguments
  parameters = inspect.signature(f).parameters

  arguments = []
  for name, par in parameters.items():
    if par.kind is inspect.Parameter.POSITIONAL_OR_KEYWORD:
      arguments.append(par.name)

    elif par.kind is inspect.Parameter.KEYWORD_ONLY:
      arguments.append(par.name)

    elif par.kind is inspect.Parameter.VAR_POSITIONAL:
      raise errors.BrainPyError('Step function do not support positional parameters, e.g., *args')
    elif par.kind is inspect.Parameter.POSITIONAL_ONLY:
      raise errors.BrainPyError('Step function do not support positional only parameters, e.g., /')
    elif par.kind is inspect.Parameter.VAR_KEYWORD:
      raise errors.BrainPyError(f'Step function do not support dict of keyword arguments: {str(par)}')
    else:
      raise errors.BrainPyError(f'Unknown argument type: {par.kind}')

  # 2. check the function arguments
  class_kw = None
  if len(arguments) > 0 and arguments[0] in profile.CLASS_KEYWORDS:
    class_kw = arguments[0]
    arguments = arguments[1:]
  for a in arguments:
    if a in profile.CLASS_KEYWORDS:
      raise errors.DiffEqError(f'Class keywords "{a}" must be defined '
                               f'as the first argument.')
  return class_kw, arguments



# -*- coding: utf-8 -*-

import jax.numpy as jnp
import numpy as np
from jax.tree_util import tree_map

from brainpy.math.jax.jaxarray import JaxArray
from brainpy.math.numpy import ops
from brainpy.tools import copy_doc

__all__ = [
  # math funcs
  'real', 'imag', 'conj', 'conjugate', 'ndim', 'isreal', 'isscalar',
  'add', 'reciprocal', 'negative', 'positive', 'multiply', 'divide',
  'power', 'subtract', 'true_divide', 'floor_divide', 'float_power',
  'fmod', 'mod', 'modf', 'divmod', 'remainder', 'abs', 'exp', 'exp2',
  'expm1', 'log', 'log10', 'log1p', 'log2', 'logaddexp', 'logaddexp2',
  'lcm', 'gcd', 'arccos', 'arccosh', 'arcsin', 'arcsinh', 'arctan',
  'arctan2', 'arctanh', 'cos', 'cosh', 'sin', 'sinc', 'sinh', 'tan',
  'tanh', 'deg2rad', 'hypot', 'rad2deg', 'degrees', 'radians', 'round',
  'around', 'round_', 'rint', 'floor', 'ceil', 'trunc', 'fix', 'prod',
  'sum', 'diff', 'median', 'nancumprod', 'nancumsum', 'nanprod', 'nansum',
  'cumprod', 'cumsum', 'ediff1d', 'cross', 'trapz', 'isfinite', 'isinf',
  'isnan', 'signbit', 'copysign', 'nextafter', 'ldexp', 'frexp', 'convolve',
  'sqrt', 'cbrt', 'square', 'absolute', 'fabs', 'sign', 'heaviside',
  'maximum', 'minimum', 'fmax', 'fmin', 'interp', 'clip', 'angle',

  # Elementwise bit operations
  'bitwise_and', 'bitwise_not', 'bitwise_or', 'bitwise_xor',
  'invert', 'left_shift', 'right_shift',

  # logic funcs
  'equal', 'not_equal', 'greater', 'greater_equal', 'less', 'less_equal',
  'array_equal', 'isclose', 'allclose', 'logical_and', 'logical_not',
  'logical_or', 'logical_xor', 'all', 'any',

  # array manipulation
  'shape', 'size', 'reshape', 'ravel', 'moveaxis', 'transpose', 'swapaxes',
  'concatenate', 'stack', 'vstack', 'hstack', 'dstack', 'column_stack',
  'split', 'dsplit', 'hsplit', 'vsplit', 'tile', 'repeat', 'unique',
  'append', 'flip', 'fliplr', 'flipud', 'roll', 'atleast_1d', 'atleast_2d',
  'atleast_3d', 'expand_dims', 'squeeze', 'sort', 'argsort', 'argmax', 'argmin',
  'argwhere', 'nonzero', 'flatnonzero', 'where', 'searchsorted', 'extract',
  'count_nonzero', 'max', 'min',

  # array creation
  'empty', 'empty_like', 'ones', 'ones_like', 'zeros', 'zeros_like', 'full',
  'full_like', 'eye', 'identity', 'array', 'asarray', 'arange', 'linspace',
  'logspace', 'meshgrid', 'diag', 'tri', 'tril', 'triu', 'vander', 'fill_diagonal',

  # indexing funcs
  'nonzero', 'where', 'tril_indices', 'tril_indices_from', 'triu_indices',
  'triu_indices_from', 'take', 'diag', 'select',

  # statistic funcs
  'nanmin', 'nanmax', 'ptp', 'percentile', 'nanpercentile', 'quantile', 'nanquantile',
  'median', 'average', 'mean', 'std', 'var', 'nanmedian', 'nanmean', 'nanstd', 'nanvar',
  'corrcoef', 'correlate', 'cov', 'histogram', 'bincount', 'digitize',

  # window funcs
  'bartlett', 'blackman', 'hamming', 'hanning', 'kaiser',

  # constants
  'e', 'pi', 'inf',

  # linear algebra
  'dot', 'vdot', 'inner', 'outer', 'kron', 'matmul', 'trace',

  # data types
  'dtype', 'finfo', 'iinfo', 'bool_', 'uint8', 'uint16', 'uint32', 'uint64',
  'int_', 'int8', 'int16', 'int32', 'int64', 'float_', 'float16', 'float32',
  'float64', 'complex_', 'complex64', 'complex128', 'set_int_', 'set_float_', 'set_complex_',

  # others
  'take_along_axis', 'clip_by_norm',
]


# math funcs
# ----------

# 1. Basics

def isreal(x):
  x = x.value if isinstance(x, JaxArray) else x
  return jnp.isreal(x)


def isscalar(x):
  x = x.value if isinstance(x, JaxArray) else x
  return jnp.isscalar(x)


def real(x):
  return x.real


def imag(x):
  return x.imag


def conj(x):
  return x.conj()


def conjugate(x):
  return x.conjugate()


def ndim(x):
  x = x.value if isinstance(x, JaxArray) else x
  return jnp.ndim(x)


# 2. Arithmetic operations

def add(x, y):
  return x + y


def reciprocal(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.reciprocal(x))


def negative(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.negative(x))


def positive(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.positive(x.value))


def multiply(x1, x2):
  if isinstance(x1, JaxArray): x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.multiply(x1, x2))


def divide(x1, x2):
  if isinstance(x1, JaxArray): x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.divide(x1, x2))


def power(x1, x2):
  if isinstance(x1, JaxArray): x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.power(x1, x2))


def subtract(x1, x2):
  if isinstance(x1, JaxArray): x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.subtract(x1, x2))


def true_divide(x1, x2):
  if isinstance(x1, JaxArray): x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.true_divide(x1, x2))


def floor_divide(x1, x2):
  if isinstance(x1, JaxArray): x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.floor_divide(x1, x2))


def float_power(x1, x2):
  if isinstance(x1, JaxArray): x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.float_power(x1, x2))


def fmod(x1, x2):
  if isinstance(x1, JaxArray): x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.fmod(x1, x2))


def mod(x1, x2):
  if isinstance(x1, JaxArray):  x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.mod(x1, x2))


def divmod(x1, x2):
  if isinstance(x1, JaxArray): x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.divmod(x1, x2))


def remainder(x1, x2):
  if isinstance(x1, JaxArray): x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.remainder(x1, x2))


def modf(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.modf(x))


def abs(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.absolute(x))


def absolute(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.absolute(x))


# 3. Exponents and logarithms
def exp(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.exp(x))


def exp2(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.exp2(x))


def expm1(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.expm1(x))


def log(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.log(x))


def log10(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.log10(x))


def log1p(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.log1p(x))


def log2(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.log2(x))


def logaddexp(x1, x2):
  if isinstance(x1, JaxArray): x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.logaddexp(x1, x2))


def logaddexp2(x1, x2):
  if isinstance(x1, JaxArray): x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.logaddexp2(x1, x2))


# 4. Rational routines
def lcm(x1, x2):
  if isinstance(x1, JaxArray): x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.lcm(x1, x2))


def gcd(x1, x2):
  if isinstance(x1, JaxArray): x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.gcd(x1, x2))


# 5. trigonometric functions

def arccos(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.arccos(x))


def arccosh(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.arccosh(x))


def arcsin(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.arcsin(x))


def arcsinh(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.arcsinh(x))


def arctan(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.arctan(x))


def arctan2(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.arctan2(x))


def arctanh(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.arctanh(x))


def cos(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.cos(x))


def cosh(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.cosh(x))


def sin(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.sin(x))


def sinc(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.sinc(x))


def sinh(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.sinh(x))


def tan(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.tan(x))


def tanh(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.tanh(x))


def deg2rad(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.deg2rad(x))


def rad2deg(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.rad2deg(x))


def degrees(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.degrees(x))


def radians(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.radians(x))


def hypot(x1, x2):
  if isinstance(x1, JaxArray): x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.hypot(x1, x2))


# 6. Rounding

def round(a, decimals=0):
  if isinstance(a, JaxArray):
    a = a.value
  return JaxArray(jnp.round(a, decimals=decimals))


around = round
round_ = round


def rint(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.rint(x))


def floor(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.floor(x))


def ceil(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.ceil(x))


def trunc(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.trunc(x))


def fix(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.fix(x))


# 7. Sums, products, differences, Reductions


def prod(a, axis=None, dtype=None, keepdims=None, initial=None, where=None):
  if isinstance(a, JaxArray): a = a.value
  r = jnp.prod(a, axis=axis, dtype=dtype, keepdims=keepdims, initial=initial, where=where)
  return r if axis is None else JaxArray(r)


def sum(a, axis=None, dtype=None, keepdims=None, initial=None, where=None):
  if isinstance(a, JaxArray): a = a.value
  r = jnp.sum(a, axis=axis, dtype=dtype, keepdims=keepdims, initial=initial, where=where)
  return r if axis is None else JaxArray(r)


def diff(a, n=1, axis: int = -1, prepend=None, append=None):
  if isinstance(a, JaxArray): a = a.value
  return JaxArray(jnp.diff(a, n=n, axis=axis, prepend=prepend, append=append))


def median(a, axis=None, keepdims=False):
  if isinstance(a, JaxArray): a = a.value
  r = jnp.median(a, axis=axis, keepdims=keepdims)
  return r if axis is None else JaxArray(r)


def nancumprod(a, axis=None, dtype=None):
  if isinstance(a, JaxArray): a = a.value
  return JaxArray(jnp.nancumprod(a=a, axis=axis, dtype=dtype))


def nancumsum(a, axis=None, dtype=None):
  if isinstance(a, JaxArray): a = a.value
  return JaxArray(jnp.nancumsum(a=a, axis=axis, dtype=dtype))


def cumprod(a, axis=None, dtype=None):
  if isinstance(a, JaxArray): a = a.value
  return JaxArray(jnp.cumprod(a=a, axis=axis, dtype=dtype))


def cumsum(a, axis=None, dtype=None):
  if isinstance(a, JaxArray): a = a.value
  return JaxArray(jnp.cumsum(a=a, axis=axis, dtype=dtype))


def nanprod(a, axis=None, dtype=None, keepdims=None):
  if isinstance(a, JaxArray): a = a.value
  r = jnp.nanprod(a=a, axis=axis, dtype=dtype, keepdims=keepdims)
  return r if axis is None else JaxArray(r)


def nansum(a, axis=None, dtype=None, keepdims=None):
  if isinstance(a, JaxArray): a = a.value
  r = jnp.nansum(a=a, axis=axis, dtype=dtype, keepdims=keepdims)
  return r if axis is None else JaxArray(r)


def ediff1d(a, to_end=None, to_begin=None):
  if isinstance(a, JaxArray): a = a.value
  if isinstance(to_end, JaxArray): to_end = to_end.value
  if isinstance(to_begin, JaxArray): to_begin = to_begin.value
  return JaxArray(jnp.ediff1d(a, to_end=to_end, to_begin=to_begin))


def cross(a, b, axisa=-1, axisb=-1, axisc=-1, axis=None):
  if isinstance(a, JaxArray): a = a.value
  if isinstance(b, JaxArray): b = b.value
  return JaxArray(jnp.cross(a, b, axisa=axisa, axisb=axisb, axisc=axisc, axis=axis))


def trapz(y, x=None, dx=1.0, axis: int = -1):
  if isinstance(y, JaxArray): y = y.value
  if isinstance(x, JaxArray): x = x.value
  return jnp.trapz(y, x=x, dx=dx, axis=axis)


# 8. floating_functions

def isfinite(x):
  if isinstance(x, JaxArray):
    return JaxArray(jnp.isfinite(x.value))
  else:
    return jnp.isfinite(x)


def isinf(x):
  if isinstance(x, JaxArray):
    return JaxArray(jnp.isinf(x.value))
  else:
    return jnp.isinf(x)


def isnan(x):
  if isinstance(x, JaxArray):
    return JaxArray(jnp.isnan(x.value))
  else:
    return jnp.isnan(x)


def signbit(x):
  if isinstance(x, JaxArray):
    return JaxArray(jnp.signbit(x.value))
  else:
    return jnp.signbit(x)


def nextafter(x1, x2):
  if isinstance(x1, JaxArray): x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.nextafter(x1, x2))


def copysign(x1, x2):
  if isinstance(x1, JaxArray): x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.copysign(x1, x2))


def ldexp(x1, x2):
  if isinstance(x1, JaxArray): x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.ldexp(x1, x2))


def frexp(x):
  if isinstance(x, JaxArray): x = x.value
  mantissa, exponent = jnp.frexp(x)
  return JaxArray(mantissa), JaxArray(exponent)


# 9. Miscellaneous

def convolve(a, v, mode='full'):
  if isinstance(a, JaxArray): a = a.value
  if isinstance(v, JaxArray): v = v.value
  return JaxArray(jnp.convolve(a, v, mode))


def sqrt(x):
  if isinstance(x, JaxArray):
    return JaxArray(jnp.sqrt(x.value))
  else:
    return jnp.sqrt(x)


def cbrt(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.cbrt(x))


def square(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.square(x))


def fabs(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.fabs(x))


def sign(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.sign(x))


def heaviside(x1, x2):
  if isinstance(x1, JaxArray): x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.heaviside(x1, x2))


def maximum(x1, x2):
  if isinstance(x1, JaxArray): x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.maximum(x1, x2))


def minimum(x1, x2):
  if isinstance(x1, JaxArray): x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.minimum(x1, x2))


def fmax(x1, x2):
  if isinstance(x1, JaxArray): x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.fmax(x1, x2))


def fmin(x1, x2):
  if isinstance(x1, JaxArray): x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.fmin(x1, x2))


def interp(x, xp, fp, left=None, right=None, period=None):
  if isinstance(x, JaxArray): x = x.value
  if isinstance(xp, JaxArray): xp = xp.value
  if isinstance(fp, JaxArray): fp = fp.value
  return JaxArray(jnp.interp(x, xp, fp, left=left, right=right, period=period))


def clip(a, a_min=None, a_max=None):
  if isinstance(a, JaxArray): a = a.value
  if isinstance(a_min, JaxArray): a_min = a_min.value
  if isinstance(a_max, JaxArray): a_max = a_max.value
  return JaxArray(jnp.clip(a, a_min, a_max))


def angle(z, deg=False):
  if isinstance(z, JaxArray): z = z.value
  a = jnp.angle(z)
  if deg:
    a *= 180 / pi
  return JaxArray(a)


# binary funcs
# -------------


def bitwise_not(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.bitwise_not(x))


def invert(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.invert(x))


def bitwise_and(x1, x2):
  if isinstance(x1, JaxArray): x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.bitwise_and(x1, x2))


def bitwise_or(x1, x2):
  if isinstance(x1, JaxArray): x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.bitwise_or(x1, x2))


def bitwise_xor(x1, x2):
  if isinstance(x1, JaxArray): x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.bitwise_xor(x1, x2))


def left_shift(x1, x2):
  if isinstance(x1, JaxArray): x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.left_shift(x1, x2))


def right_shift(x1, x2):
  if isinstance(x1, JaxArray): x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.right_shift(x1, x2))


# logic funcs
# -----------

# 1. Comparison

def equal(x1, x2):
  if isinstance(x1, JaxArray): x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.equal(x1, x2))


def not_equal(x1, x2):
  if isinstance(x1, JaxArray): x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.not_equal(x1, x2))


def greater(x1, x2):
  if isinstance(x1, JaxArray): x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.greater(x1, x2))


def greater_equal(x1, x2):
  if isinstance(x1, JaxArray): x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.greater_equal(x1, x2))


def less(x1, x2):
  if isinstance(x1, JaxArray): x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.less(x1, x2))


def less_equal(x1, x2):
  if isinstance(x1, JaxArray): x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.less_equal(x1, x2))


def array_equal(a, b, equal_nan=False):
  if isinstance(a, JaxArray): a = a.value
  if isinstance(b, JaxArray): b = b.value
  return jnp.array_equal(a, b, equal_nan=equal_nan)


def isclose(a, b, rtol=1e-05, atol=1e-08, equal_nan=False):
  if isinstance(a, JaxArray): a = a.value
  if isinstance(b, JaxArray): b = b.value
  return JaxArray(jnp.isclose(a, b, rtol=rtol, atol=atol, equal_nan=equal_nan))


def allclose(a, b, rtol=1e-05, atol=1e-08, equal_nan=False):
  if isinstance(a, JaxArray): a = a.value
  if isinstance(b, JaxArray): b = b.value
  return jnp.allclose(a, b, rtol=rtol, atol=atol, equal_nan=equal_nan)


# 2. Logical operations
def logical_not(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.logical_not(x))


def logical_and(x1, x2):
  if isinstance(x1, JaxArray): x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.logical_and(x1, x2))


def logical_or(x1, x2):
  if isinstance(x1, JaxArray): x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.logical_or(x1, x2))


def logical_xor(x1, x2):
  if isinstance(x1, JaxArray): x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.logical_xor(x1, x2))


# 3. Truth value testing

def all(a, axis=None, keepdims=None, where=None):
  if isinstance(a, JaxArray): a = a.value
  r = jnp.all(a=a, axis=axis, keepdims=keepdims, where=where)
  return r if axis is None else JaxArray(r)


def any(a, axis=None, keepdims=None, where=None):
  if isinstance(a, JaxArray): a = a.value
  r = jnp.any(a=a, axis=axis, keepdims=keepdims, where=where)
  return r if axis is None else JaxArray(r)


# array manipulation
# ------------------


def shape(x):
  if isinstance(x, JaxArray): x = x.value
  return jnp.shape(x)


def size(x, axis=None):
  if isinstance(x, JaxArray): x = x.value
  r = jnp.size(x, axis=axis)
  return r if axis is None else JaxArray(r)


def reshape(x, newshape, order="C"):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.reshape(x, newshape, order=order))


def ravel(x, order="C"):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.ravel(x, order=order))


def moveaxis(x, source, destination):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.moveaxis(x, source, destination))


def transpose(x, axis=None):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.transpose(x, axes=axis))


def swapaxes(x, axis1, axis2):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.swapaxes(x, axis1, axis2))


def concatenate(arrays, axis: int = 0):
  arrays = [a.value if isinstance(a, JaxArray) else a for a in arrays]
  return JaxArray(jnp.concatenate(arrays, axis))


def stack(arrays, axis: int = 0):
  arrays = [a.value if isinstance(a, JaxArray) else a for a in arrays]
  return JaxArray(jnp.stack(arrays, axis))


def vstack(arrays):
  arrays = [a.value if isinstance(a, JaxArray) else a for a in arrays]
  return JaxArray(jnp.vstack(arrays))


def hstack(arrays):
  arrays = [a.value if isinstance(a, JaxArray) else a for a in arrays]
  return JaxArray(jnp.hstack(arrays))


def dstack(arrays):
  arrays = [a.value if isinstance(a, JaxArray) else a for a in arrays]
  return JaxArray(jnp.dstack(arrays))


def column_stack(arrays):
  arrays = [a.value if isinstance(a, JaxArray) else a for a in arrays]
  return JaxArray(jnp.column_stack(arrays))


def split(ary, indices_or_sections, axis=0):
  if isinstance(ary, JaxArray): ary = ary.value
  if isinstance(indices_or_sections, JaxArray): indices_or_sections = indices_or_sections.value
  return [JaxArray(a) for a in jnp.split(ary, indices_or_sections, axis=axis)]


def dsplit(ary, indices_or_sections):
  return split(ary, indices_or_sections, axis=2)


def hsplit(ary, indices_or_sections):
  return split(ary, indices_or_sections, axis=1)


def vsplit(ary, indices_or_sections):
  return split(ary, indices_or_sections, axis=0)


def tile(A, reps):
  if isinstance(A, JaxArray): A = A.value
  return JaxArray(jnp.tile(A, reps))


def repeat(x, repeats, axis=None):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.repeat(x, repeats=repeats, axis=axis))


def unique(x, return_index=False, return_inverse=False,
           return_counts=False, axis=None):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.unique(x,
                             return_index=return_index,
                             return_inverse=return_inverse,
                             return_counts=return_counts,
                             axis=axis))


def append(arr, values, axis=None):
  if isinstance(arr, JaxArray): arr = arr.value
  if isinstance(values, JaxArray): values = values.value
  return JaxArray(jnp.append(arr, values, axis=axis))


def flip(x, axis=None):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.flip(x, axis=axis))


def fliplr(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.fliplr(x))


def flipud(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.flipud(x))


def roll(x, shift, axis=None):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.roll(x, shift, axis=axis))


def atleast_1d(*arys):
  return jnp.atleast_1d(*[a.value if isinstance(a, JaxArray) else a for a in arys])


def atleast_2d(*arys):
  return jnp.atleast_2d(*[a.value if isinstance(a, JaxArray) else a for a in arys])


def atleast_3d(*arys):
  return jnp.atleast_3d(*[a.value if isinstance(a, JaxArray) else a for a in arys])


def expand_dims(x, axis):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.expand_dims(x, axis=axis))


def squeeze(x, axis=None):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.squeeze(x, axis=axis))


def sort(x, axis=-1, kind='quicksort', order=None):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.sort(x, axis=axis, kind=kind, order=order))


def argsort(x, axis=-1, kind='quicksort', order=None):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.argsort(x, axis=axis, kind=kind, order=order))


def argmax(x, axis=None):
  if isinstance(x, JaxArray): x = x.value
  r = jnp.argmax(x, axis=axis)
  return r if axis is None else JaxArray(r)


def argmin(x, axis=None):
  if isinstance(x, JaxArray): x = x.value
  r = jnp.argmin(x, axis=axis)
  return r if axis is None else JaxArray(r)


def argwhere(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.argwhere(x))


def nonzero(x):
  if isinstance(x, JaxArray): x = x.value
  return jnp.nonzero(x)


def flatnonzero(x):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.flatnonzero(x))


def where(condition, x=None, y=None):
  if isinstance(condition, JaxArray): condition = condition.value
  if isinstance(x, JaxArray): x = x.value
  if isinstance(y, JaxArray): y = y.value
  return jnp.where(condition, x=x, y=y)


def searchsorted(a, v, side='left', sorter=None):
  if isinstance(a, JaxArray): a = a.value
  if isinstance(v, JaxArray): v = v.value
  return JaxArray(jnp.searchsorted(a, v, side=side, sorter=sorter))


def extract(condition, arr):
  if isinstance(condition, JaxArray): condition = condition.value
  if isinstance(arr, JaxArray): arr = arr.value
  return JaxArray(jnp.extract(condition, arr))


def count_nonzero(a, axis=None, keepdims=False):
  if isinstance(a, JaxArray): a = a.value
  return jnp.count_nonzero(a, axis=axis, keepdims=keepdims)


def max(a, axis=None, keepdims=None, initial=None, where=None):
  if isinstance(a, JaxArray): a = a.value
  r = jnp.max(a, axis=axis, keepdims=keepdims, initial=initial, where=where)
  return r if axis is None else JaxArray(r)


def min(a, axis=None, keepdims=None, initial=None, where=None):
  if isinstance(a, JaxArray): a = a.value
  r = jnp.min(a, axis=axis, keepdims=keepdims, initial=initial, where=where)
  return r if axis is None else JaxArray(r)


# array creation
# --------------

def zeros(shape, dtype=None):
  return JaxArray(jnp.zeros(shape, dtype=dtype))


def ones(shape, dtype=None):
  return JaxArray(jnp.ones(shape, dtype=dtype))


def full(shape, fill_value, dtype=None):
  return JaxArray(jnp.full(shape, fill_value, dtype=dtype))


def empty(shape, dtype=None):
  return JaxArray(jnp.zeros(shape, dtype=dtype))


def zeros_like(a, dtype=None, shape=None):
  if isinstance(a, JaxArray): a = a.value
  return JaxArray(jnp.zeros_like(a, dtype=dtype, shape=shape))


def ones_like(a, dtype=None, shape=None):
  if isinstance(a, JaxArray): a = a.value
  return JaxArray(jnp.ones_like(a, dtype=dtype, shape=shape))


def empty_like(a, dtype=None, shape=None):
  if isinstance(a, JaxArray): a = a.value
  return JaxArray(jnp.zeros_like(a, dtype=dtype, shape=shape))


def full_like(a, fill_value, dtype=None, shape=None):
  if isinstance(a, JaxArray): a = a.value
  return JaxArray(jnp.full_like(a, fill_value, dtype=dtype, shape=shape))


def eye(N, M=None, k=0, dtype=None):
  return JaxArray(jnp.eye(N, M=M, k=k, dtype=dtype))


def identity(n, dtype=None):
  return JaxArray(jnp.identity(n, dtype=dtype))


def array(object, dtype=None, copy=True, order="K", ndmin=0):
  if isinstance(object, JaxArray): object = object.value
  return JaxArray(jnp.array(object, dtype=dtype, copy=copy, order=order, ndmin=ndmin))


def asarray(a, dtype=None, order=None):
  if isinstance(a, JaxArray): a = a.value
  return JaxArray(jnp.asarray(a=a, dtype=dtype, order=order))


def arange(*args, **kwargs):
  return JaxArray(jnp.arange(*args, **kwargs))


def linspace(*args, **kwargs):
  return JaxArray(jnp.linspace(*args, **kwargs))


def logspace(*args, **kwargs):
  return JaxArray(jnp.logspace(*args, **kwargs))


def meshgrid(*xi, copy=True, sparse=False, indexing='xy'):
  xi = [x.value if isinstance(x, JaxArray) else x for x in xi]
  return JaxArray(jnp.meshgrid(*xi, copy=copy, sparse=sparse, indexing=indexing))


def diag(a, k=0):
  if isinstance(a, JaxArray): a = a.value
  return JaxArray(jnp.diag(a, k))


def tri(N, M=None, k=0, dtype=None):
  return JaxArray(jnp.tri(N, M=M, k=k, dtype=dtype))


def tril(a, k=0):
  if isinstance(a, JaxArray): a = a.value
  return JaxArray(jnp.tril(a, k))


def triu(a, k=0):
  if isinstance(a, JaxArray): a = a.value
  return JaxArray(jnp.triu(a, k))


def vander(x, N=None, increasing=False):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.vander(x, N=N, increasing=increasing))


def fill_diagonal(a, val):
  if isinstance(a, JaxArray): a = a.value
  assert a.ndim >= 2
  i, j = jnp.diag_indices(min(a.shape[-2:]))
  return JaxArray(a.at[..., i, j].set(val))


# indexing funcs
# --------------

tril_indices = jnp.tril_indices
triu_indices = jnp.triu_indices


def tril_indices_from(x, k=0):
  if isinstance(x, JaxArray): x = x.value
  return jnp.tril_indices_from(x, k=k)


def triu_indices_from(x, k=0):
  if isinstance(x, JaxArray): x = x.value
  return jnp.triu_indices_from(x, k=k)


def take(x, indices, axis=None, mode=None):
  if isinstance(x, JaxArray): x = x.value
  if isinstance(indices, JaxArray): indices = indices.value
  return JaxArray(jnp.take(x, indices=indices, axis=axis, mode=mode))


def select(condlist, choicelist, default=0):
  condlist = [c.value if isinstance(c, JaxArray) else c for c in condlist]
  choicelist = [c.value if isinstance(c, JaxArray) else c for c in choicelist]
  return JaxArray(jnp.select(condlist, choicelist, default=default))


# statistic funcs
# ---------------

def nanmin(x, axis=None, keepdims=None):
  if isinstance(x, JaxArray): x = x.value
  r = jnp.nanmin(x, axis=axis, keepdims=keepdims)
  return r if axis is None else JaxArray(r)


def nanmax(x, axis=None, keepdims=None):
  if isinstance(x, JaxArray): x = x.value
  r = jnp.nanmax(x, axis=axis, keepdims=keepdims)
  return r if axis is None else JaxArray(r)


def ptp(x, axis=None, keepdims=None):
  if isinstance(x, JaxArray): x = x.value
  r = jnp.ptp(x, axis=axis, keepdims=keepdims)
  return r if axis is None else JaxArray(r)


def percentile(a, q, axis=None, interpolation='linear', keepdims=False):
  if isinstance(a, JaxArray): a = a.value
  if isinstance(q, JaxArray): q = q.value
  r = jnp.percentile(a=a, q=q, axis=axis, interpolation=interpolation, keepdims=keepdims)
  return r if axis is None else JaxArray(r)


def nanpercentile(a, q, axis=None, interpolation='linear', keepdims=False):
  if isinstance(a, JaxArray): a = a.value
  if isinstance(q, JaxArray): q = q.value
  r = jnp.nanpercentile(a=a, q=q, axis=axis, interpolation=interpolation, keepdims=keepdims)
  return r if axis is None else JaxArray(r)


def quantile(a, q, axis=None, interpolation='linear', keepdims=False):
  if isinstance(a, JaxArray): a = a.value
  if isinstance(q, JaxArray): q = q.value
  r = jnp.quantile(a=a, q=q, axis=axis, interpolation=interpolation, keepdims=keepdims)
  return r if axis is None else JaxArray(r)


def nanquantile(a, q, axis=None, interpolation='linear', keepdims=False):
  if isinstance(a, JaxArray): a = a.value
  if isinstance(q, JaxArray): q = q.value
  r = jnp.nanquantile(a=a, q=q, axis=axis, interpolation=interpolation, keepdims=keepdims)
  return r if axis is None else JaxArray(r)


def average(a, axis=None, weights=None, returned=False):
  if isinstance(a, JaxArray): a = a.value
  if isinstance(weights, JaxArray): weights = weights.value
  r = jnp.average(a, axis=axis, weights=weights, returned=returned)
  return r if axis is None else JaxArray(r)


def mean(a, axis=None, dtype=None, keepdims=None, where=None):
  if isinstance(a, JaxArray): a = a.value
  r = jnp.mean(a, axis=axis, dtype=dtype, keepdims=keepdims, where=where)
  return r if axis is None else JaxArray(r)


def std(a, axis=None, dtype=None, ddof=0, keepdims=None, where=None):
  if isinstance(a, JaxArray): a = a.value
  r = jnp.std(a=a, axis=axis, dtype=dtype, ddof=ddof, keepdims=keepdims, where=where)
  return r if axis is None else JaxArray(r)


def var(a, axis=None, dtype=None, ddof=0, keepdims=None, where=None):
  if isinstance(a, JaxArray): a = a.value
  r = jnp.var(a, axis=axis, dtype=dtype, ddof=ddof, keepdims=keepdims, where=where)
  return r if axis is None else JaxArray(r)


def nanmedian(a, axis=None, keepdims=False):
  return nanquantile(a, 0.5, axis=axis, keepdims=keepdims, interpolation='midpoint')


def nanmean(a, axis=None, dtype=None, keepdims=None):
  if isinstance(a, JaxArray): a = a.value
  r = jnp.nanmean(a, axis=axis, dtype=dtype, keepdims=keepdims)
  return r if axis is None else JaxArray(r)


def nanstd(a, axis=None, dtype=None, ddof=0, keepdims=None, where=None):
  if isinstance(a, JaxArray): a = a.value
  r = jnp.nanstd(a=a, axis=axis, dtype=dtype, ddof=ddof, keepdims=keepdims, where=where)
  return r if axis is None else JaxArray(r)


def nanvar(a, axis=None, dtype=None, ddof=0, keepdims=None, where=None):
  if isinstance(a, JaxArray): a = a.value
  r = jnp.nanvar(a, axis=axis, dtype=dtype, ddof=ddof, keepdims=keepdims, where=where)
  return r if axis is None else JaxArray(r)


def corrcoef(x, y=None, rowvar=True):
  if isinstance(x, JaxArray): x = x.value
  if isinstance(y, JaxArray): y = y.value
  return JaxArray(jnp.corrcoef(x, y, rowvar))


def correlate(a, v, mode='valid'):
  if isinstance(a, JaxArray): a = a.value
  if isinstance(v, JaxArray): v = v.value
  return JaxArray(jnp.correlate(a, v, mode))


def cov(m, y=None, rowvar=True, bias=False, ddof=None, fweights=None, aweights=None):
  if isinstance(m, JaxArray): m = m.value
  if isinstance(y, JaxArray): y = y.value
  if isinstance(fweights, JaxArray): fweights = fweights.value
  if isinstance(aweights, JaxArray): aweights = aweights.value
  return JaxArray(jnp.cov(m, y=y, rowvar=rowvar, bias=bias, ddof=ddof,
                          fweights=fweights, aweights=aweights))


def histogram(a, bins=10, range=None, weights=None, density=None):
  if isinstance(a, JaxArray): a = a.value
  if isinstance(weights, JaxArray): weights = weights.value
  hist, bin_edges = jnp.histogram(a=a, bins=bins, range=range, weights=weights, density=density)
  return JaxArray(hist), JaxArray(bin_edges)


def bincount(x, weights=None, minlength=None):
  if isinstance(x, JaxArray): x = x.value
  if isinstance(weights, JaxArray): weights = weights.value
  return JaxArray(jnp.bincount(x, weights=weights, minlength=minlength))


def digitize(x, bins, right=False):
  if isinstance(x, JaxArray): x = x.value
  if isinstance(bins, JaxArray): bins = bins.value
  return JaxArray(jnp.digitize(x, bins=bins, right=right))


def bartlett(M):
  return JaxArray(jnp.bartlett(M))


def blackman(M):
  return JaxArray(jnp.blackman(M))


def hamming(M):
  return JaxArray(jnp.hamming(M))


def hanning(M):
  return JaxArray(jnp.hanning(M))


def kaiser(M, beta):
  return JaxArray(jnp.kaiser(M, beta))


# constants
# ---------

e = jnp.e
pi = jnp.pi
inf = jnp.inf


# linear algebra
# --------------


def dot(x1, x2):
  if isinstance(x1, JaxArray): x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.dot(x1, x2))


def vdot(x1, x2):
  if isinstance(x1, JaxArray): x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.vdot(x1, x2))


def inner(x1, x2):
  if isinstance(x1, JaxArray): x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.inner(x1, x2))


def outer(x1, x2):
  if isinstance(x1, JaxArray): x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.outer(x1, x2))


def kron(x1, x2):
  if isinstance(x1, JaxArray): x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.kron(x1, x2))


def matmul(x1, x2):
  if isinstance(x1, JaxArray): x1 = x1.value
  if isinstance(x2, JaxArray): x2 = x2.value
  return JaxArray(jnp.matmul(x1, x2))


def trace(x, offset=0, axis1=0, axis2=1, dtype=None):
  if isinstance(x, JaxArray): x = x.value
  return JaxArray(jnp.trace(x, offset=offset, axis1=axis1, axis2=axis2, dtype=dtype))


# data types
# ----------

dtype = jnp.dtype
finfo = jnp.finfo
iinfo = jnp.iinfo

bool_ = jnp.bool_
uint8 = jnp.uint8
uint16 = jnp.uint16
uint32 = jnp.uint32
uint64 = jnp.uint64
int_ = jnp.int_
int8 = jnp.int8
int16 = jnp.int16
int32 = jnp.int32
int64 = jnp.int64
float_ = jnp.float_
float16 = jnp.float16
float32 = jnp.float32
float64 = jnp.float64
complex_ = jnp.complex_
complex64 = jnp.complex64
complex128 = jnp.complex128


def set_int_(int_type):
  global int_
  assert isinstance(int_type, type)
  int_ = int_type


def set_float_(float_type):
  global float_
  assert isinstance(float_type, type)
  float_ = float_type


def set_complex_(complex_type):
  global complex_
  assert isinstance(complex_type, type)
  complex_ = complex_type


# others
# ------
@copy_doc(np.take_along_axis)
def take_along_axis(a, indices, axis):
  if isinstance(a, JaxArray): a = a.value
  if isinstance(indices, JaxArray): indices = indices.value
  return JaxArray(jnp.take_along_axis(a, indices, axis))


@copy_doc(ops.clip_by_norm)
def clip_by_norm(t, clip_norm, axis=None):
  f = lambda l: l * clip_norm / maximum(sqrt(sum(l * l, axis=axis, keepdims=True)), clip_norm)
  return tree_map(f, t)



# -*- coding: utf-8 -*-


from brainpy import math, tools
from .base import TwoEndConnector


__all__ = [
  'MatConn',
  'IJConn',
]


class MatConn(TwoEndConnector):
  """Connector built from the connection matrix."""
  def __init__(self, conn_mat):
    super(MatConn, self).__init__()

    assert isinstance(conn_mat, math.ndarray) and conn_mat.ndim == 2
    self.conn_mat = math.asarray(conn_mat, dtype=math.bool_)
    self.pre_num, self.post_num = conn_mat.shape

  def __call__(self, pre_size, post_size):
    if isinstance(pre_size, int): pre_size = (pre_size,)
    pre_size = tuple(pre_size)
    if isinstance(post_size, int): post_size = (post_size,)
    post_size = tuple(post_size)
    self.pre_size, self.post_size = pre_size, post_size
    assert self.pre_num == tools.size2num(pre_size)
    assert self.post_num == tools.size2num(post_size)
    return self

  def require(self, structures):
    self.check(structures)
    return self.returns(mat=self.conn_mat)


class IJConn(TwoEndConnector):
  """Connector built from the ``pre_ids`` and ``post_ids`` connections."""
  def __init__(self, i, j):
    super(IJConn, self).__init__()

    assert isinstance(i, math.ndarray) and i.ndim == 1
    assert isinstance(j, math.ndarray) and j.ndim == 1
    assert i.size == j.size

    # initialize the class via "pre_ids" and "post_ids"
    self.pre_ids = math.asarray(i, dtype=math.int_)
    self.post_ids = math.asarray(j, dtype=math.int_)

  def require(self, structures):
    self.check(structures)
    return self.returns(ij=(self.pre_ids, self.post_ids))



# -*- coding: utf-8 -*-

import logging

from brainpy import errors

__all__ = [
  'grad', 'value_and_grad',
]

logger = logging.getLogger('brainpy.math.numpy.gradient')


def grad(func, *args, **kwargs):
  _msg = '"grad" is only supported in JAX backend, not available in numpy backend. \n' \
         'You can switch to JAX backend by `brainpy.math.use_backend("jax")`'
  logger.error(_msg)
  raise errors.BrainPyError(_msg)


def value_and_grad(func, *args, **kwargs):
  _msg = '"value_and_grad" is only supported in JAX backend, not available in numpy backend. \n' \
         'You can switch to JAX backend by `brainpy.math.use_backend("jax")`'
  logger.error(_msg)
  raise errors.BrainPyError(_msg)



# -*- coding: utf-8 -*-

import brainpy as bp
from pprint import pprint


class HH(bp.NeuGroup):
  def __init__(self, size, ENa=55., EK=-90., EL=-65, C=1.0,
               gNa=35., gK=9., gL=0.1, V_th=20., phi=5.0, **kwargs):
    super(HH, self).__init__(size=size, **kwargs)

    # parameters
    self.ENa = ENa
    self.EK = EK
    self.EL = EL
    self.C = C
    self.gNa = gNa
    self.gK = gK
    self.gL = gL
    self.V_th = V_th
    self.phi = phi

    # variables
    self.V = bp.math.ones(self.num) * -65.
    self.h = bp.math.ones(self.num) * 0.6
    self.n = bp.math.ones(self.num) * 0.32
    self.inputs = bp.math.zeros(self.num)
    self.spikes = bp.math.zeros(self.num, dtype=bp.math.bool_)

    self.pre = None

  @bp.odeint
  def integral(self, V, h, n, t, Iext):
    alpha = 0.07 * bp.math.exp(-(V + 58) / 20)
    beta = 1 / (bp.math.exp(-0.1 * (V + 28)) + 1)
    dhdt = self.phi * (alpha * (1 - h) - beta * h)

    alpha = -0.01 * (V + 34) / (bp.math.exp(-0.1 * (V + 34)) - 1)
    beta = 0.125 * bp.math.exp(-(V + 44) / 80)
    dndt = self.phi * (alpha * (1 - n) - beta * n)

    m_alpha = -0.1 * (V + 35) / (bp.math.exp(-0.1 * (V + 35)) - 1)
    m_beta = 4 * bp.math.exp(-(V + 60) / 18)
    m = m_alpha / (m_alpha + m_beta)
    INa = self.gNa * m ** 3 * h * (V - self.ENa)
    IK = self.gK * n ** 4 * (V - self.EK)
    IL = self.gL * (V - self.EL)
    dVdt = (- INa - IK - IL + Iext) / self.C

    return dVdt, dhdt, dndt

  def update(self, _t, _i):
    V, h, n = self.integral(self.V, self.h, self.n, _t, self.inputs)
    self.spikes[:] = bp.math.logical_and(self.V < self.V_th, V >= self.V_th)
    self.V[:] = V
    self.h[:] = h
    self.n[:] = n
    self.inputs[:] = 0.


def test_nodes():
  A = HH(1, name='X')
  B = HH(1, name='Y')
  A.pre = B
  B.pre = A

  net = bp.Network(A, B)
  abs_nodes = net.nodes(method='absolute')
  rel_nodes = net.nodes(method='relative')
  print()
  pprint(abs_nodes)
  pprint(rel_nodes)

  assert len(abs_nodes) == 3
  assert len(rel_nodes) == 5


def test_ints():
  A = HH(1, name='X2')
  B = HH(1, name='Y2')
  A.pre = B
  B.pre = A

  net = bp.Network(A, B)
  abs_ints = net.ints(method='absolute')
  rel_ints = net.ints(method='relative')
  print()
  pprint(abs_ints.keys())
  pprint(rel_ints.keys())




# -*- coding: utf-8 -*-

import numpy as np

from brainpy import math

__all__ = [
  'section_input',
  'constant_input', 'constant_current',
  'spike_input', 'spike_current',
  'ramp_input', 'ramp_current',
]


def section_input(values, durations, dt=None, return_length=False):
  """Format an input current with different sections.

  For example:

  If you want to get an input where the size is 0 bwteen 0-100 ms,
  and the size is 1. between 100-200 ms.
  >>> import numpy as np
  >>> section_input(values=[0, 1],
  >>>               durations=[100, 100])

  Parameters
  ----------
  values : list, np.ndarray
      The current values for each period duration.
  durations : list, np.ndarray
      The duration for each period.
  dt : float
      Default is None.
  return_length : bool
      Return the final duration length.

  Returns
  -------
  current_and_duration : tuple
      (The formatted current, total duration)
  """
  assert len(durations) == len(values), f'"values" and "durations" must be the same length, while ' \
                                        f'we got {len(values)} != {len(durations)}.'

  dt = math.get_dt() if dt is None else dt

  # get input current shape, and duration
  I_duration = sum(durations)
  I_shape = ()
  for val in values:
    shape = math.shape(val)
    if len(shape) > len(I_shape):
      I_shape = shape

  # get the current
  start = 0
  I_current = math.zeros((int(np.ceil(I_duration / dt)),) + I_shape, dtype=math.float_)
  for c_size, duration in zip(values, durations):
    length = int(duration / dt)
    I_current[start: start + length] = c_size
    start += length

  if return_length:
    return I_current, I_duration
  else:
    return I_current


def constant_input(I_and_duration, dt=None):
  """Format constant input in durations.

  For example:

  If you want to get an input where the size is 0 bwteen 0-100 ms,
  and the size is 1. between 100-200 ms.
  >>> import numpy as np
  >>> constant_input([(0, 100), (1, 100)])
  >>> constant_input([(np.zeros(100), 100), (np.random.rand(100), 100)])

  Parameters
  ----------
  I_and_duration : list
      This parameter receives the current size and the current
      duration pairs, like `[(Isize1, duration1), (Isize2, duration2)]`.
  dt : float
      Default is None.

  Returns
  -------
  current_and_duration : tuple
      (The formatted current, total duration)
  """
  dt = math.get_dt() if dt is None else dt

  # get input current dimension, shape, and duration
  I_duration = 0.
  I_shape = ()
  for I in I_and_duration:
    I_duration += I[1]
    shape = math.shape(I[0])
    if len(shape) > len(I_shape):
      I_shape = shape

  # get the current
  start = 0
  I_current = math.zeros((int(np.ceil(I_duration / dt)),) + I_shape, dtype=math.float_)
  for c_size, duration in I_and_duration:
    length = int(duration / dt)
    I_current[start: start + length] = c_size
    start += length
  return I_current, I_duration


constant_current = constant_input


def spike_input(sp_times, sp_lens, sp_sizes, duration, dt=None):
  """Format current input like a series of short-time spikes.

  For example:

  If you want to generate a spike train at 10 ms, 20 ms, 30 ms, 200 ms, 300 ms,
  and each spike lasts 1 ms and the spike current is 0.5, then you can use the
  following funtions:

  >>> spike_input(sp_times=[10, 20, 30, 200, 300],
  >>>             sp_lens=1.,  # can be a list to specify the spike length at each point
  >>>             sp_sizes=0.5,  # can be a list to specify the current size at each point
  >>>             duration=400.)

  Parameters
  ----------
  sp_times : list, tuple
      The spike time-points. Must be an iterable object.
  sp_lens : int, float, list, tuple
      The length of each point-current, mimicking the spike durations.
  sp_sizes : int, float, list, tuple
      The current sizes.
  duration : int, float
      The total current duration.
  dt : float
      The default is None.

  Returns
  -------
  current : math.ndarray
      The formatted input current.
  """
  dt = math.get_dt() if dt is None else dt
  assert isinstance(sp_times, (list, tuple))
  if isinstance(sp_lens, (float, int)):
    sp_lens = [sp_lens] * len(sp_times)
  if isinstance(sp_sizes, (float, int)):
    sp_sizes = [sp_sizes] * len(sp_times)

  current = math.zeros(int(np.ceil(duration / dt)), dtype=math.float_)
  for time, dur, size in zip(sp_times, sp_lens, sp_sizes):
    pp = int(time / dt)
    p_len = int(dur / dt)
    current[pp: pp + p_len] = size
  return current


spike_current = spike_input


def ramp_input(c_start, c_end, duration, t_start=0, t_end=None, dt=None):
  """Get the gradually changed input current.

  Parameters
  ----------
  c_start : float
      The minimum (or maximum) current size.
  c_end : float
      The maximum (or minimum) current size.
  duration : int, float
      The total duration.
  t_start : float
      The ramped current start time-point.
  t_end : float
      The ramped current end time-point. Default is the None.
  dt : float, int, optional
      The numerical precision.

  Returns
  -------
  current_and_duration : tuple
      (The formatted current, total duration)
  """
  dt = math.get_dt() if dt is None else dt
  t_end = duration if t_end is None else t_end

  current = math.zeros(int(np.ceil(duration / dt)), dtype=math.float_)
  p1 = int(np.ceil(t_start / dt))
  p2 = int(np.ceil(t_end / dt))
  current[p1: p2] = math.array(math.linspace(c_start, c_end, p2 - p1), dtype=math.float_)
  return current


ramp_current = ramp_input




# -*- coding: utf-8 -*-

from brainpy import math
from brainpy.simulation.brainobjects.base import DynamicalSystem

__all__ = [
  'BatchNorm',
]


class BatchNorm(DynamicalSystem):
  """Batch Normalization module.

  Normalizes inputs to maintain a mean of ~0 and stddev of ~1 [1]_.

  Parameters
  ----------
  axis : int
  momentum : float
    The value used to compute exponential moving average of batch statistics.
  eps : float
    The small value which is used for numerical stability.

  References
  ----------
  .. [1] Li, Xiang, et al. “Understanding the Disharmony Between Dropout and
         Batch Normalization by Variance Shift.” 2019 IEEE/CVF Conference on
         Computer Vision and Pattern Recognition (CVPR), 2019, pp. 2682–2690.

  """

  def __init__(self, dims, axis, momentum=0.99, eps=1e-6, name=None):
    super(BatchNorm, self).__init__(name=name)

    self.axis = axis
    self.momentum = momentum
    self.eps = eps
    self.running_mean = math.zeros(dims)
    self.running_var = math.zeros(dims)
    self.shift = math.TrainVar(math.zeros(dims))
    self.scale = math.TrainVar(math.ones(dims))

  def update(self, x, **kwargs):
    if kwargs.get('train', True):
      pass
    else:
      pass



# -*- coding: utf-8 -*-

from pprint import pprint

from brainpy.integrators import constants, utils

_SDE_UNKNOWN_NO = 0


def basic_info(f, g):
  vdt = 'dt'
  if f.__name__.isidentifier():
    func_name = f.__name__
  elif g.__name__.isidentifier():
    func_name = g.__name__
  else:
    global _SDE_UNKNOWN_NO
    func_name = f'unknown_sde{_SDE_UNKNOWN_NO}'
  func_new_name = constants.SDE_INT + func_name
  class_kw, variables, parameters, arguments = utils.get_args(f)
  return vdt, variables, parameters, arguments, func_new_name


def compile_and_assign_attrs(code_lines, code_scope, show_code,
                             variables, parameters, func_name,
                             intg_type, var_type, wiener_type, dt,
                             method, raw_func):
  code_scope_old = {key: val for key, val in code_scope.items()}

  # compile functions
  code = '\n'.join(code_lines)
  if show_code:
    print(code)
    print()
    pprint(code_scope)
    print()
  exec(compile(code, '', 'exec'), code_scope)
  new_f = code_scope[func_name]

  # assign values
  new_f.brainpy_data = dict(raw_func=raw_func,
                            code_lines=code_lines,
                            code_scope=code_scope_old,
                            variables=variables,
                            parameters=parameters,
                            dt=dt,
                            func_name=func_name,
                            var_type=var_type,
                            intg_type=intg_type,
                            wiener_type=wiener_type,
                            method=method)
  return new_f



# -*- coding: utf-8 -*-

import gc
import logging

import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.mplot3d import Axes3D

from brainpy import errors, math
from brainpy.analysis import stability
from brainpy.analysis.symbolic import base, utils
from brainpy.analysis.symbolic.trajectory import Trajectory


logger = logging.getLogger('brainpy.analysis.symbolic')


__all__ = [
  'Bifurcation',
  '_Bifurcation1D',
  '_Bifurcation2D',

  'FastSlowBifurcation',
  '_FastSlow1D',
  '_FastSlow2D',
]


class Bifurcation(object):
  """A tool class for bifurcation analysis.

  The bifurcation analyzer is restricted to analyze the bifurcation
  relation between membrane potential and a given model parameter
  (co-dimension-1 case) or two model parameters (co-dimension-2 case).

  Externally injected current is also treated as a model parameter in
  this class, instead of a model state.

  Parameters
  ----------

  integrals : function, functions
      The integral functions defined with `brainpy.odeint` or
      `brainpy.sdeint` or `brainpy.ddeint`, or `brainpy.fdeint`.
  target_vars : dict
      The target dynamical variables. It must a dictionary which
      specifies the boundary of the variables: `{'var1': [min, max]}`.
  fixed_vars : dict, optional
      The fixed variables. It must a fixed value with the format of `{'var1': value}`.
  target_pars : dict, optional
      The parameters which can be dynamical varied. It must be a dictionary which
      specifies the boundary of the variables: `{'par1': [min, max]}`
  pars_update : dict, optional
      The parameters to update. Or, they can be treated as staitic parameters.
      Same with the `fixed_vars`, they are must fixed values with the format of
      `{'par1': value}`.
  numerical_resolution : float, dict, optional
      The resolution for numerical iterative solvers. Default is 0.1.
      It can set the numerical resolution of dynamical variables or dynamical parameters.
      For example, set ``numerical_resolution=0.1`` will generalize it to all
      variables and parameters;
      set ``numerical_resolution={var1: 0.1, var2: 0.2, par1: 0.1, par2: 0.05}`` will
      specify the particular resolutions to variables and parameters.
      Moreover, you can also set
      ``numerical_resolution={var1: np.array([...]), var2: 0.1}`` to specify the
      search points need to explore for variable `var1`. This will be useful to
      set sense search points at some inflection points.
  options : dict, optional
      The other setting parameters, which includes:

          perturbation
              float. The small perturbation used to solve the function derivatives.
          sympy_solver_timeout
              float, with the unit of second. The maximum time allowed to use sympy solver
              to get the variable relationship.
          escape_sympy_solver
              bool. Whether escape to use sympy solver, and directly use numerical optimization
              method to solve the nullcline and fixed points.
          lim_scale
              float. The axis limit scale factor. Default is 1.05. The setting means
              the axes will be clipped to ``[var_min * (1-lim_scale)/2, var_max * (var_max-1)/2]``.

      The parameters which are usefull for two-dimensional bifurcation analysis:

          shgo_args
              dict. Arguments of `shgo` optimization method, which can be used to set the
              fields of: constraints, n, iters, callback, minimizer_kwargs, options,
              sampling_method.
          show_shgo
              bool. whether print the shgo's value.
          fl_tol
              float. The tolerance of the function value to recognize it as a candidate of
              function root point.
          xl_tol
              float. The tolerance of the l2 norm distances between this point and previous
              points. If the norm distances are all bigger than `xl_tol` means this
              point belong to a new function root point.
  """

  def __init__(self, integrals, target_pars, target_vars, fixed_vars=None, pars_update=None,
               numerical_resolution=0.1, options=None):
    # check "model"
    self.model = utils.transform_integrals_to_model(integrals)

    # check "target_pars"
    if not isinstance(target_pars, dict):
      raise errors.BrainPyError('"target_pars" must a dict with the format of: '
                                 '{"Parameter A": [A_min, A_max],'
                                 ' "Parameter B": [B_min, B_max]}')
    self.target_pars = target_pars
    if len(target_pars) > 2:
      raise errors.BrainPyError("The number of parameters in bifurcation"
                                 "analysis cannot exceed 2.")

    # check "fixed_vars"
    if fixed_vars is None:
      fixed_vars = dict()
    if not isinstance(fixed_vars, dict):
      raise errors.BrainPyError('"fixed_vars" must be a dict the format of: '
                                 '{"Variable A": A_value, "Variable B": B_value}')
    self.fixed_vars = fixed_vars

    # check "target_vars"
    if not isinstance(target_vars, dict):
      raise errors.BrainPyError('"target_vars" must a dict with the format of: '
                                 '{"Variable A": [A_min, A_max], "Variable B": [B_min, B_max]}')
    self.target_vars = target_vars

    # check "pars_update"
    if pars_update is None:
      pars_update = dict()
    if not isinstance(pars_update, dict):
      raise errors.BrainPyError('"pars_update" must be a dict the format of: '
                                 '{"Par A": A_value, "Par B": B_value}')
    for key in pars_update.keys():
      if (key not in self.model.scopes) and (key not in self.model.parameters):
        raise errors.BrainPyError(f'"{key}" is not a valid parameter in "{integrals}". ')
    self.pars_update = pars_update

    # bifurcation analysis
    if len(self.target_vars) == 1:
      self.analyzer = _Bifurcation1D(model_or_integrals=self.model,
                                     target_pars=target_pars,
                                     target_vars=target_vars,
                                     fixed_vars=fixed_vars,
                                     pars_update=pars_update,
                                     numerical_resolution=numerical_resolution,
                                     options=options)

    elif len(self.target_vars) == 2:
      self.analyzer = _Bifurcation2D(model_or_integrals=self.model,
                                     target_pars=target_pars,
                                     target_vars=target_vars,
                                     fixed_vars=fixed_vars,
                                     pars_update=pars_update,
                                     numerical_resolution=numerical_resolution,
                                     options=options)

    else:
      raise errors.BrainPyError(f'Cannot analyze three dimensional system: {self.target_vars}')

  def plot_bifurcation(self, *args, **kwargs):
    """Plot bifurcation, which support bifurcation analysis of
    co-dimension 1 and co-dimension 2.

    Parameters
    ----------
    show : bool
        Whether show the bifurcation figure.

    Returns
    -------
    points : dict
        The bifurcation points which specifies their fixed points
        and corresponding stability.
    """
    return self.analyzer.plot_bifurcation(*args, **kwargs)

  def plot_limit_cycle_by_sim(self, var, duration=100, inputs=(), plot_style=None, tol=0.001, show=False):
    """Plot limit cycles by the simulation results.

    This function help users plot the limit cycles through the simulation results,
    in which the periodic signals will be automatically found and then treated them
    as the candidate of limit cycles.

    Parameters
    ----------
    var : str
        The target variable to found its limit cycles.
    duration : int, float, tuple, list
        The simulation duration.
    inputs : tuple, list
        The simulation inputs.
    plot_style : dict
        The limit cycle plotting style settings.
    tol : float
        The tolerance to found periodic signals.
    show : bool
        Whether show the figure.
    """
    self.analyzer.plot_limit_cycle_by_sim(var=var, duration=duration, inputs=inputs,
                                          plot_style=plot_style, tol=tol, show=show)


class _Bifurcation1D(base.Base1DAnalyzer):
  """Bifurcation analysis of 1D system.

  Using this class, we can make co-dimension1 or co-dimension2 bifurcation analysis.
  """

  def __init__(self, model_or_integrals, target_pars, target_vars, fixed_vars=None,
               pars_update=None, numerical_resolution=0.1, options=None):
    super(_Bifurcation1D, self).__init__(model_or_integrals=model_or_integrals,
                                         target_pars=target_pars,
                                         target_vars=target_vars,
                                         fixed_vars=fixed_vars,
                                         pars_update=pars_update,
                                         numerical_resolution=numerical_resolution,
                                         options=options)

  def plot_bifurcation(self, show=False):
    logger.info('plot bifurcation ...')

    f_fixed_point = self.get_f_fixed_point()
    f_dfdx = self.get_f_dfdx()

    if len(self.target_pars) == 1:
      container = {c: {'p': [], 'x': []} for c in stability.get_1d_stability_types()}

      # fixed point
      par_a = self.dpar_names[0]
      for p in self.resolutions[par_a]:
        xs = f_fixed_point(p)
        for x in xs:
          dfdx = f_dfdx(x, p)
          fp_type = stability.stability_analysis(dfdx)
          container[fp_type]['p'].append(p)
          container[fp_type]['x'].append(x)

      # visualization
      plt.figure(self.x_var)
      for fp_type, points in container.items():
        if len(points['x']):
          plot_style = stability.plot_scheme[fp_type]
          plt.plot(points['p'], points['x'], '.', **plot_style, label=fp_type)
      plt.xlabel(par_a)
      plt.ylabel(self.x_var)

      # scale = (self.options.lim_scale - 1) / 2
      # plt.xlim(*utils.rescale(self.target_pars[self.dpar_names[0]], scale=scale))
      # plt.ylim(*utils.rescale(self.target_vars[self.x_var], scale=scale))

      plt.legend()
      if show:
        plt.show()

    elif len(self.target_pars) == 2:
      container = {c: {'p0': [], 'p1': [], 'x': []} for c in stability.get_1d_stability_types()}

      # fixed point
      for p0 in self.resolutions[self.dpar_names[0]]:
        for p1 in self.resolutions[self.dpar_names[1]]:
          xs = f_fixed_point(p0, p1)
          for x in xs:
            dfdx = f_dfdx(x, p0, p1)
            fp_type = stability.stability_analysis(dfdx)
            container[fp_type]['p0'].append(p0)
            container[fp_type]['p1'].append(p1)
            container[fp_type]['x'].append(x)

      # visualization
      fig = plt.figure(self.x_var)
      ax = fig.add_subplot(projection='3d')
      for fp_type, points in container.items():
        if len(points['x']):
          plot_style = stability.plot_scheme[fp_type]
          xs = points['p0']
          ys = points['p1']
          zs = points['x']
          ax.scatter(xs, ys, zs, **plot_style, label=fp_type)

      ax.set_xlabel(self.dpar_names[0])
      ax.set_ylabel(self.dpar_names[1])
      ax.set_zlabel(self.x_var)

      # scale = (self.options.lim_scale - 1) / 2
      # ax.set_xlim(*utils.rescale(self.target_pars[self.dpar_names[0]], scale=scale))
      # ax.set_ylim(*utils.rescale(self.target_pars[self.dpar_names[1]], scale=scale))
      # ax.set_zlim(*utils.rescale(self.target_vars[self.x_var], scale=scale))

      ax.grid(True)
      ax.legend()
      if show:
        plt.show()

    else:
      raise errors.BrainPyError(f'Cannot visualize co-dimension {len(self.target_pars)} '
                                 f'bifurcation.')
    return container

  def plot_limit_cycle_by_sim(self, *args, **kwargs):
    raise NotImplementedError('1D phase plane do not support plot_limit_cycle_by_sim.')


class _Bifurcation2D(base.Base2DAnalyzer):
  """Bifurcation analysis of 2D system.

  Using this class, we can make co-dimension1 or co-dimension2 bifurcation analysis.
  """

  def __init__(self, model_or_integrals, target_pars, target_vars, fixed_vars=None,
               pars_update=None, numerical_resolution=0.1, options=None):
    super(_Bifurcation2D, self).__init__(model_or_integrals=model_or_integrals,
                                         target_pars=target_pars,
                                         target_vars=target_vars,
                                         fixed_vars=fixed_vars,
                                         pars_update=pars_update,
                                         numerical_resolution=numerical_resolution,
                                         options=options)

    self.fixed_points = None

  def plot_bifurcation(self, show=False):
    logger.info('plot bifurcation ...')

    # functions
    f_fixed_point = self.get_f_fixed_point()
    f_jacobian = self.get_f_jacobian()

    # bifurcation analysis of co-dimension 1
    if len(self.target_pars) == 1:
      container = {c: {'p': [], self.x_var: [], self.y_var: []}
                   for c in stability.get_2d_stability_types()}

      # fixed point
      for p in self.resolutions[self.dpar_names[0]]:
        xs, ys = f_fixed_point(p)
        for x, y in zip(xs, ys):
          dfdx = f_jacobian(x, y, p)
          fp_type = stability.stability_analysis(dfdx)
          container[fp_type]['p'].append(p)
          container[fp_type][self.x_var].append(x)
          container[fp_type][self.y_var].append(y)

      # visualization
      for var in self.dvar_names:
        plt.figure(var)
        for fp_type, points in container.items():
          if len(points['p']):
            plot_style = stability.plot_scheme[fp_type]
            plt.plot(points['p'], points[var], '.', **plot_style, label=fp_type)
        plt.xlabel(self.dpar_names[0])
        plt.ylabel(var)

        # scale = (self.options.lim_scale - 1) / 2
        # plt.xlim(*utils.rescale(self.target_pars[self.dpar_names[0]], scale=scale))
        # plt.ylim(*utils.rescale(self.target_vars[var], scale=scale))

        plt.legend()
      if show:
        plt.show()

    # bifurcation analysis of co-dimension 2
    elif len(self.target_pars) == 2:
      container = {c: {'p0': [], 'p1': [], self.x_var: [], self.y_var: []}
                   for c in stability.get_2d_stability_types()}

      # fixed point
      for p0 in self.resolutions[self.dpar_names[0]]:
        for p1 in self.resolutions[self.dpar_names[1]]:
          xs, ys = f_fixed_point(p0, p1)
          for x, y in zip(xs, ys):
            dfdx = f_jacobian(x, y, p0, p1)
            fp_type = stability.stability_analysis(dfdx)
            container[fp_type]['p0'].append(p0)
            container[fp_type]['p1'].append(p1)
            container[fp_type][self.x_var].append(x)
            container[fp_type][self.y_var].append(y)

      # visualization
      for var in self.dvar_names:
        fig = plt.figure(var)
        ax = fig.add_subplot(projection='3d')
        for fp_type, points in container.items():
          if len(points['p0']):
            plot_style = stability.plot_scheme[fp_type]
            xs = points['p0']
            ys = points['p1']
            zs = points[var]
            ax.scatter(xs, ys, zs, **plot_style, label=fp_type)

        ax.set_xlabel(self.dpar_names[0])
        ax.set_ylabel(self.dpar_names[1])
        ax.set_zlabel(var)

        # scale = (self.options.lim_scale - 1) / 2
        # ax.set_xlim(*utils.rescale(self.target_pars[self.dpar_names[0]], scale=scale))
        # ax.set_ylim(*utils.rescale(self.target_pars[self.dpar_names[1]], scale=scale))
        # ax.set_zlim(*utils.rescale(self.target_vars[var], scale=scale))

        ax.grid(True)
        ax.legend()
      if show:
        plt.show()

    else:
      raise ValueError('Unknown length of parameters.')

    self.fixed_points = container
    return container

  def plot_limit_cycle_by_sim(self, var, duration=100, inputs=(), plot_style=None, tol=0.001, show=False):
    logger.info('plot limit cycle ...')

    if self.fixed_points is None:
      raise errors.AnalyzerError('Please call "plot_bifurcation()" before "plot_limit_cycle_by_sim()".')
    if plot_style is None:
      plot_style = dict()
    fmt = plot_style.pop('fmt', '.')

    if var not in [self.x_var, self.y_var]:
      raise errors.AnalyzerError()

    all_xs, all_ys, all_p0, all_p1 = [], [], [], []

    # unstable node
    unstable_node = self.fixed_points[stability.UNSTABLE_NODE_2D]
    all_xs.extend(unstable_node[self.x_var])
    all_ys.extend(unstable_node[self.y_var])
    if len(self.dpar_names) == 1:
      all_p0.extend(unstable_node['p'])
    elif len(self.dpar_names) == 2:
      all_p0.extend(unstable_node['p0'])
      all_p1.extend(unstable_node['p1'])
    else:
      raise ValueError

    # unstable focus
    unstable_focus = self.fixed_points[stability.UNSTABLE_FOCUS_2D]
    all_xs.extend(unstable_focus[self.x_var])
    all_ys.extend(unstable_focus[self.y_var])
    if len(self.dpar_names) == 1:
      all_p0.extend(unstable_focus['p'])
    elif len(self.dpar_names) == 2:
      all_p0.extend(unstable_focus['p0'])
      all_p1.extend(unstable_focus['p1'])
    else:
      raise ValueError

    # format points
    all_xs = np.array(all_xs)
    all_ys = np.array(all_ys)
    all_p0 = np.array(all_p0)
    all_p1 = np.array(all_p1)

    # fixed variables
    fixed_vars = dict()
    for key, val in self.fixed_vars.items():
      fixed_vars[key] = val
    fixed_vars[self.dpar_names[0]] = all_p0
    if len(self.dpar_names) == 2:
      fixed_vars[self.dpar_names[1]] = all_p1

    # initialize neuron group
    length = all_xs.shape[0]
    traj_group = Trajectory(size=length,
                            integrals=self.model.integrals,
                            target_vars={self.x_var: all_xs, self.y_var: all_ys},
                            fixed_vars=fixed_vars,
                            pars_update=self.pars_update,
                            scope=self.model.scopes)
    traj_group.run(duration=duration)

    # find limit cycles
    limit_cycle_max = []
    limit_cycle_min = []
    # limit_cycle = []
    p0_limit_cycle = []
    p1_limit_cycle = []
    for i in range(length):
      data = traj_group.mon[var][:, i]
      max_index = utils.find_indexes_of_limit_cycle_max(data, tol=tol)
      if max_index[0] != -1:
        x_cycle = data[max_index[0]: max_index[1]]
        limit_cycle_max.append(data[max_index[1]])
        limit_cycle_min.append(x_cycle.min())
        # limit_cycle.append(x_cycle)
        p0_limit_cycle.append(all_p0[i])
        if len(self.dpar_names) == 2:
          p1_limit_cycle.append(all_p1[i])
    self.fixed_points['limit_cycle'] = {var: {'max': limit_cycle_max,
                                              'min': limit_cycle_min,
                                              # 'cycle': limit_cycle
                                              }}
    p0_limit_cycle = np.array(p0_limit_cycle)
    p1_limit_cycle = np.array(p1_limit_cycle)

    # visualization
    if len(self.dpar_names) == 2:
      self.fixed_points['limit_cycle'] = {'p0': p0_limit_cycle, 'p1': p1_limit_cycle}
      plt.figure(var)
      plt.plot(p0_limit_cycle, p1_limit_cycle, limit_cycle_max, **plot_style, label='limit cycle (max)')
      plt.plot(p0_limit_cycle, p1_limit_cycle, limit_cycle_min, **plot_style, label='limit cycle (min)')
      plt.legend()

    else:
      self.fixed_points['limit_cycle'] = {'p': p0_limit_cycle}
      if len(limit_cycle_max):
        plt.figure(var)
        plt.plot(p0_limit_cycle, limit_cycle_max, fmt, **plot_style, label='limit cycle (max)')
        plt.plot(p0_limit_cycle, limit_cycle_min, fmt, **plot_style, label='limit cycle (min)')
        plt.legend()

    if show:
      plt.show()

    del traj_group
    gc.collect()


class FastSlowBifurcation(object):
  """Fast slow analysis analysis proposed by John Rinzel [1]_ [2]_ [3]_.

  (J Rinzel, 1985, 1986, 1987) proposed that in a fast-slow dynamical
  system, we can treat the slow variables as the bifurcation parameters,
  and then study how the different value of slow variables affect the
  bifurcation of the fast sub-system.

  Parameters
  ----------

  integrals : function, functions
      The integral functions defined with `brainpy.odeint` or
      `brainpy.sdeint` or `brainpy.ddeint`, or `brainpy.fdeint`.
  fast_vars : dict
      The fast dynamical variables. It must a dictionary which
      specifies the boundary of the variables: `{'var1': [min, max]}`.
  slow_vars : dict
      The slow dynamical variables. It must a dictionary which
      specifies the boundary of the variables: `{'var1': [min, max]}`.
  fixed_vars : dict
      The fixed variables. It must a fixed value with the format of `{'var1': value}`.
  pars_update : dict, optional
      The parameters to update. Or, they can be treated as staitic parameters.
      Same with the `fixed_vars`, they are must fixed values with the format of
      `{'par1': value}`.
  numerical_resolution : float, dict
      The resolution for numerical iterative solvers. Default is 0.1.
      It can set the numerical resolution of dynamical variables or dynamical parameters.
      For example, set ``numerical_resolution=0.1`` will generalize it to all
      variables and parameters;
      set ``numerical_resolution={var1: 0.1, var2: 0.2, par1: 0.1, par2: 0.05}`` will
      specify the particular resolutions to variables and parameters.
      Moreover, you can also set
      ``numerical_resolution={var1: np.array([...]), var2: 0.1}`` to specify the
      search points need to explore for variable `var1`. This will be useful to
      set sense search points at some inflection points.
  options : dict, optional
      The other setting parameters, which includes:

          perturbation
              float. The small perturbation used to solve the function derivatives.
          sympy_solver_timeout
              float, with the unit of second. The maximum time allowed to use sympy solver
              to get the variable relationship.
          escape_sympy_solver
              bool. Whether escape to use sympy solver, and directly use numerical optimization
              method to solve the nullcline and fixed points.
          lim_scale
              float. The axis limit scale factor. Default is 1.05. The setting means
              the axes will be clipped to ``[var_min * (1-lim_scale)/2, var_max * (var_max-1)/2]``.

  References
  ----------

  .. [1] Rinzel, John. "Bursting oscillations in an excitable
         membrane model." In Ordinary and partial differential
         equations, pp. 304-316. Springer, Berlin, Heidelberg, 1985.
  .. [2] Rinzel, John , and Y. S. Lee . On Different Mechanisms for
         Membrane Potential Bursting. Nonlinear Oscillations in
         Biology and Chemistry. Springer Berlin Heidelberg, 1986.
  .. [3] Rinzel, John. "A formal classification of bursting mechanisms
         in excitable systems." In Mathematical topics in population
         biology, morphogenesis and neurosciences, pp. 267-281.
         Springer, Berlin, Heidelberg, 1987.

  """

  def __init__(self, integrals, fast_vars, slow_vars, fixed_vars=None,
               pars_update=None, numerical_resolution=0.1, options=None):
    # check "model"
    self.model = utils.transform_integrals_to_model(integrals)

    # check "fast_vars"
    if not isinstance(fast_vars, dict):
      raise errors.BrainPyError('"fast_vars" must a dict with the format of: '
                                 '{"Var A": [A_min, A_max],'
                                 ' "Var B": [B_min, B_max]}')
    self.fast_vars = fast_vars
    if len(fast_vars) > 2:
      raise errors.BrainPyError("FastSlowBifurcation can only analyze the system with less "
                                 "than two-variable fast subsystem.")

    # check "slow_vars"
    if not isinstance(slow_vars, dict):
      raise errors.BrainPyError('"slow_vars" must a dict with the format of: '
                                 '{"Variable A": [A_min, A_max], '
                                 '"Variable B": [B_min, B_max]}')
    self.slow_vars = slow_vars
    if len(slow_vars) > 2:
      raise errors.BrainPyError("FastSlowBifurcation can only analyze the system with less "
                                 "than two-variable slow subsystem.")
    for key in self.slow_vars:
      self.model.variables.remove(key)
      self.model.parameters.append(key)

    # check "fixed_vars"
    if fixed_vars is None:
      fixed_vars = dict()
    if not isinstance(fixed_vars, dict):
      raise errors.BrainPyError('"fixed_vars" must be a dict the format of: '
                                 '{"Variable A": A_value, "Variable B": B_value}')
    self.fixed_vars = fixed_vars

    # check "pars_update"
    if pars_update is None:
      pars_update = dict()
    if not isinstance(pars_update, dict):
      raise errors.BrainPyError('"pars_update" must be a dict the format of: '
                                 '{"Par A": A_value, "Par B": B_value}')
    for key in pars_update.keys():
      if (key not in self.model.scopes) and (key not in self.model.parameters):
        raise errors.BrainPyError(f'"{key}" is not a valid parameter in "{integrals}" model. ')
    self.pars_update = pars_update

    # bifurcation analysis
    if len(self.fast_vars) == 1:
      self.analyzer = _FastSlow1D(model_or_integrals=self.model,
                                  fast_vars=fast_vars,
                                  slow_vars=slow_vars,
                                  fixed_vars=fixed_vars,
                                  pars_update=pars_update,
                                  numerical_resolution=numerical_resolution,
                                  options=options)

    elif len(self.fast_vars) == 2:
      self.analyzer = _FastSlow2D(model_or_integrals=self.model,
                                  fast_vars=fast_vars,
                                  slow_vars=slow_vars,
                                  fixed_vars=fixed_vars,
                                  pars_update=pars_update,
                                  numerical_resolution=numerical_resolution,
                                  options=options)

    else:
      raise errors.BrainPyError(f'Cannot analyze {len(fast_vars)} dimensional fast system.')

  def plot_bifurcation(self, *args, **kwargs):
    """Plot bifurcation.

    Parameters
    ----------
    show : bool
        Whether show the bifurcation figure.

    Returns
    -------
    points : dict
        The bifurcation points which specifies their fixed points
        and corresponding stability.
    """
    return self.analyzer.plot_bifurcation(*args, **kwargs)

  def plot_trajectory(self, *args, **kwargs):
    """Plot trajectory.

    This function helps users to plot specific trajectories.

    Parameters
    ----------
    initials : list, tuple
        The initial value setting of the targets. It can be a tuple/list of floats to specify
        each value of dynamical variables (for example, ``(a, b)``). It can also be a
        tuple/list of tuple to specify multiple initial values (for example,
        ``[(a1, b1), (a2, b2)]``).
    duration : int, float, tuple, list
        The running duration. Same with the ``duration`` in ``NeuGroup.run()``.
        It can be a int/float (``t_end``) to specify the same running end time,
        or it can be a tuple/list of int/float (``(t_start, t_end)``) to specify
        the start and end simulation time. Or, it can be a list of tuple
        (``[(t1_start, t1_end), (t2_start, t2_end)]``) to specify the specific
        start and end simulation time for each initial value.
    plot_duration : tuple/list of tuple, optional
        The duration to plot. It can be a tuple with ``(start, end)``. It can
        also be a list of tuple ``[(start1, end1), (start2, end2)]`` to specify
        the plot duration for each initial value running.
    show : bool
        Whether show or not.
    """
    self.analyzer.plot_trajectory(*args, **kwargs)

  def plot_limit_cycle_by_sim(self, *args, **kwargs):
    """Plot limit cycles by the simulation results.

    This function help users plot the limit cycles through the simulation results,
    in which the periodic signals will be automatically found and then treated them
    as the candidate of limit cycles.

    Parameters
    ----------
    var : str
        The target variable to found its limit cycles.
    duration : int, float, tuple, list
        The simulation duration.
    inputs : tuple, list
        The simulation inputs.
    plot_style : dict
        The limit cycle plotting style settings.
    tol : float
        The tolerance to found periodic signals.
    show : bool
        Whether show the figure.
    """
    self.analyzer.plot_limit_cycle_by_sim(*args, **kwargs)


class _FastSlowTrajectory(object):
  def __init__(self, model_or_intgs, fast_vars, slow_vars, fixed_vars=None,
               pars_update=None, **kwargs):
    if isinstance(model_or_intgs, utils.DynamicModel):
      self.model = model_or_intgs
    elif (isinstance(model_or_intgs, (list, tuple)) and callable(model_or_intgs[0])) or callable(model_or_intgs):
      self.model = utils.transform_integrals_to_model(model_or_intgs)
    else:
      raise ValueError
    self.fast_vars = fast_vars
    self.slow_vars = slow_vars
    self.fixed_vars = fixed_vars
    self.pars_update = pars_update
    options = kwargs.get('options', dict())
    if options is None:
      options = dict()
    self.lim_scale = options.get('lim_scale', 1.05)

    # fast variables
    self.fast_var_names = list(fast_vars.keys())

    # slow variables
    self.slow_var_names = list(slow_vars.keys())

  def plot_trajectory(self, initials, duration, plot_duration=None, show=False):
    """Plot trajectories according to the settings.

    Parameters
    ----------
    initials : list, tuple
        The initial value setting of the targets. It can be a tuple/list of floats to specify
        each value of dynamical variables (for example, ``(a, b)``). It can also be a
        tuple/list of tuple to specify multiple initial values (for example,
        ``[(a1, b1), (a2, b2)]``).
    duration : int, float, tuple, list
        The running duration. Same with the ``duration`` in ``NeuGroup.run()``.
        It can be a int/float (``t_end``) to specify the same running end time,
        or it can be a tuple/list of int/float (``(t_start, t_end)``) to specify
        the start and end simulation time. Or, it can be a list of tuple
        (``[(t1_start, t1_end), (t2_start, t2_end)]``) to specify the specific
        start and end simulation time for each initial value.
    plot_duration : tuple/list of tuple, optional
        The duration to plot. It can be a tuple with ``(start, end)``. It can
        also be a list of tuple ``[(start1, end1), (start2, end2)]`` to specify
        the plot duration for each initial value running.
    show : bool
        Whether show or not.
    """
    logger.info('plot trajectory ...')

    # 1. format the initial values
    all_vars = self.fast_var_names + self.slow_var_names
    if isinstance(initials, dict):
      initials = [initials]
    elif isinstance(initials, (list, tuple)):
      if isinstance(initials[0], (int, float)):
        initials = [{all_vars[i]: v for i, v in enumerate(initials)}]
      elif isinstance(initials[0], dict):
        initials = initials
      elif isinstance(initials[0], (tuple, list)) and isinstance(initials[0][0], (int, float)):
        initials = [{all_vars[i]: v for i, v in enumerate(init)} for init in initials]
      else:
        raise ValueError
    else:
      raise ValueError
    for initial in initials:
      if len(initial) != len(all_vars):
        raise errors.AnalyzerError(f'Should provide all fast-slow variables ({all_vars}) '
                                   f' initial values, but we only get initial values for '
                                   f'variables {list(initial.keys())}.')

    # 2. format the running duration
    if isinstance(duration, (int, float)):
      duration = [(0, duration) for _ in range(len(initials))]
    elif isinstance(duration[0], (int, float)):
      duration = [duration for _ in range(len(initials))]
    else:
      assert len(duration) == len(initials)

    # 3. format the plot duration
    if plot_duration is None:
      plot_duration = duration
    if isinstance(plot_duration[0], (int, float)):
      plot_duration = [plot_duration for _ in range(len(initials))]
    else:
      assert len(plot_duration) == len(initials)

    # 5. run the network
    for init_i, initial in enumerate(initials):
      traj_group = Trajectory(size=1,
                              integrals=self.model.integrals,
                              target_vars=initial,
                              fixed_vars=self.fixed_vars,
                              pars_update=self.pars_update,
                              scope=self.model.scopes)
      traj_group.run(duration=duration[init_i], report=False)

      #   5.3 legend
      legend = f'$traj_{init_i}$: '
      for key in all_vars:
        legend += f'{key}={initial[key]}, '
      legend = legend[:-2]

      #   5.4 trajectory
      start = int(plot_duration[init_i][0] / math.get_dt())
      end = int(plot_duration[init_i][1] / math.get_dt())

      #   5.5 visualization
      for var_name in self.fast_var_names:
        s0 = traj_group.mon[self.slow_var_names[0]][start: end, 0]
        fast = traj_group.mon[var_name][start: end, 0]

        fig = plt.figure(var_name)
        if len(self.slow_var_names) == 1:
          lines = plt.plot(s0, fast, label=legend)
          utils.add_arrow(lines[0])
          # middle = int(s0.shape[0] / 2)
          # plt.arrow(s0[middle], fast[middle],
          #           s0[middle + 1] - s0[middle], fast[middle + 1] - fast[middle],
          #           shape='full')

        elif len(self.slow_var_names) == 2:
          fig.gca()
          s1 = traj_group.mon[self.slow_var_names[1]][start: end, 0]
          plt.plot(s0, s1, fast, label=legend)
        else:
          raise errors.AnalyzerError

    # 6. visualization
    for var_name in self.fast_vars.keys():
      fig = plt.figure(var_name)

      # scale = (self.lim_scale - 1.) / 2
      if len(self.slow_var_names) == 1:
        # plt.xlim(*utils.rescale(self.slow_vars[self.slow_var_names[0]], scale=scale))
        # plt.ylim(*utils.rescale(self.fast_vars[var_name], scale=scale))
        plt.xlabel(self.slow_var_names[0])
        plt.ylabel(var_name)
      elif len(self.slow_var_names) == 2:
        ax = fig.add_subplot(projection='3d')
        # ax.set_xlim(*utils.rescale(self.slow_vars[self.slow_var_names[0]], scale=scale))
        # ax.set_ylim(*utils.rescale(self.slow_vars[self.slow_var_names[1]], scale=scale))
        # ax.set_zlim(*utils.rescale(self.fast_vars[var_name], scale=scale))
        ax.set_xlabel(self.slow_var_names[0])
        ax.set_ylabel(self.slow_var_names[1])
        ax.set_zlabel(var_name)

      plt.legend()

    if show:
      plt.show()


class _FastSlow1D(_Bifurcation1D):
  def __init__(self, model_or_integrals, fast_vars, slow_vars, fixed_vars=None,
               pars_update=None, numerical_resolution=0.1, options=None):
    super(_FastSlow1D, self).__init__(model_or_integrals=model_or_integrals,
                                      target_pars=slow_vars,
                                      target_vars=fast_vars,
                                      fixed_vars=fixed_vars,
                                      pars_update=pars_update,
                                      numerical_resolution=numerical_resolution,
                                      options=options)
    self.traj = _FastSlowTrajectory(model_or_intgs=model_or_integrals,
                                    fast_vars=fast_vars,
                                    slow_vars=slow_vars,
                                    fixed_vars=fixed_vars,
                                    pars_update=pars_update,
                                    numerical_resolution=numerical_resolution,
                                    options=options)

  def plot_trajectory(self, *args, **kwargs):
    self.traj.plot_trajectory(*args, **kwargs)

  def plot_bifurcation(self, *args, **kwargs):
    return super(_FastSlow1D, self).plot_bifurcation(*args, **kwargs)

  def plot_limit_cycle_by_sim(self, *args, **kwargs):
    super(_FastSlow1D, self).plot_limit_cycle_by_sim(*args, **kwargs)


class _FastSlow2D(_Bifurcation2D):
  def __init__(self, model_or_integrals, fast_vars, slow_vars, fixed_vars=None,
               pars_update=None, numerical_resolution=0.1, options=None):
    super(_FastSlow2D, self).__init__(model_or_integrals=model_or_integrals,
                                      target_pars=slow_vars,
                                      target_vars=fast_vars,
                                      fixed_vars=fixed_vars,
                                      pars_update=pars_update,
                                      numerical_resolution=numerical_resolution,
                                      options=options)
    self.traj = _FastSlowTrajectory(model_or_intgs=model_or_integrals,
                                    fast_vars=fast_vars,
                                    slow_vars=slow_vars,
                                    fixed_vars=fixed_vars,
                                    pars_update=pars_update,
                                    numerical_resolution=numerical_resolution,
                                    options=options)

  def plot_trajectory(self, *args, **kwargs):
    self.traj.plot_trajectory(*args, **kwargs)

  def plot_bifurcation(self, *args, **kwargs):
    return super(_FastSlow2D, self).plot_bifurcation(*args, **kwargs)

  def plot_limit_cycle_by_sim(self, *args, **kwargs):
    super(_FastSlow2D, self).plot_limit_cycle_by_sim(*args, **kwargs)


if __name__ == '__main__':
  Axes3D
